#include "Challenge.hpp"
#include "Solution.hpp"
#include "Coord.hpp"
#include "Trf.hpp"
#include "Geometry.hpp"
#include "Random.hpp"

using std::string;
using std::vector;

class Facet
{
public:
    Facet() : m_dir(Dir::CCW) {}

    vector<uint32_t> m_vertices;
    Trf m_trf;
    Trf m_invTrf;
    Dir m_dir;
};

void makeChallenge(uint64_t seed,
                   Solution* pSolution)
{
    auto& solution = *pSolution;
    solution.m_sourceCoords.clear();
    solution.m_facets.clear();
    solution.m_destCoords.clear();

    Random random(seed);

    vector<Coord> sourceCoords;
    sourceCoords.push_back(Coord(0, 0));
    sourceCoords.push_back(Coord(1, 0));
    sourceCoords.push_back(Coord(1, 1));
    sourceCoords.push_back(Coord(0, 1));

    vector<Coord> destCoords;

#if 0
    Trf initTrf = Trf(1, 0,
                      0, 1,
                      0, 0);
#endif

#if 1
    Trf initTrf = Trf(mpq_class("4/5"), mpq_class("-3/5"),
                      mpq_class("3/5"), mpq_class("4/5"),
                      0, 0);
#endif

#if 0
    initTrf.x += mpq_class("135246357468147258369470/13579086422468097531");
    initTrf.y += mpq_class("333774462228833923764587/66884455993322001177");
#endif

#if 1
    initTrf.x += mpq_class("2/5");
    initTrf.y += mpq_class("1/5");
#endif

    Trf invInitTrf = invert(initTrf);

    vector<Facet> facets;
    {
        facets.emplace_back();
        auto& facet = facets.back();
        facet.m_vertices.push_back(0);
        facet.m_vertices.push_back(1);
        facet.m_vertices.push_back(2);
        facet.m_vertices.push_back(3);

        facet.m_dir = Dir::CCW;

        facet.m_trf = initTrf;
        facet.m_invTrf = invInitTrf;

        for (auto& sourceCoord : sourceCoords)
        {
            destCoords.push_back(facet.m_trf * sourceCoord);
        }
    }

    // Move coord1 to coord2
    auto connect = [&](const Coord& coord1, const Coord& coord2)
    {
        Coord lineOrigin = (coord1 + coord2) / 2;
        Coord coord2dir = coord2 - lineOrigin;
        Coord lineDir(-coord2dir.y, coord2dir.x);

        Trf mirrorTrf = getMirrorTrf(lineOrigin, lineDir);

        /*
        printf("coord1 = (%s, %s)\n",
               coord1.x.get_str().c_str(),
               coord1.y.get_str().c_str());
        printf("coord2 = (%s, %s)\n",
               coord2.x.get_str().c_str(),
               coord2.y.get_str().c_str());
        printf("mirrorTrf(coord1) = (%s, %s)\n",
               (mirrorTrf*coord1).x.get_str().c_str(),
               (mirrorTrf*coord1).y.get_str().c_str());
        */

        vector<Facet> newFacets;

        uint32_t facetCount = (uint32_t)facets.size();
        for (uint32_t iFacet = 0; iFacet < facetCount; iFacet++)
        {
            bool needFlip = false;
            bool needOrig = false;
            auto& facet = facets[iFacet];
            vector<int32_t> sides;
            for (uint32_t vertexIdx : facet.m_vertices)
            {
                Coord vec = destCoords[vertexIdx] - lineOrigin;
                int32_t side = sgn(cross(lineDir, vec));
                sides.push_back(side);
                if (side > 0)
                {
                    needFlip = true;
                }
                else if (side < 0)
                {
                    needOrig = true;
                }
            }

            if (needOrig)
            {
                newFacets.emplace_back();
                auto& newFacet = newFacets.back();
                newFacet.m_trf = facet.m_trf;
                newFacet.m_invTrf = facet.m_invTrf;
                newFacet.m_dir = facet.m_dir;

                uint32_t iPrevIdx = (uint32_t)facet.m_vertices.size() - 1;
                for (uint32_t iIdx = 0; iIdx < (uint32_t)facet.m_vertices.size(); iIdx++)
                {
                    uint32_t idx = facet.m_vertices[iIdx];
                    uint32_t prevIdx = facet.m_vertices[iPrevIdx];
                    auto& coord = destCoords[idx];
                    auto& prevCoord = destCoords[prevIdx];
                    int32_t side = sides[iIdx];
                    int32_t prevSide = sides[iPrevIdx];
                    iPrevIdx = iIdx;
                    if (side < 0 && prevSide <= 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side == 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side > 0 && prevSide >= 0)
                    {
                        continue;
                    }

                    mpq_class lineParam;
                    mpq_class segParam;
                    Coord segDir = coord - prevCoord;
                    if (!intersectLines(lineOrigin,
                                        lineDir,
                                        prevCoord,
                                        segDir,
                                        &lineParam,
                                        &segParam))
                    {
                        printf("Lines don't intersect???");
                        exit(1);
                    }

                    Coord intersection = prevCoord + segParam * segDir;
                    Coord sourceIntersection = facet.m_invTrf * intersection;
                    uint32_t intersectionIdx = 0;
                    for (; intersectionIdx < (uint32_t)sourceCoords.size(); intersectionIdx++)
                    {
                        if (sourceCoords[intersectionIdx] == sourceIntersection)
                        {
                            break;
                        }
                    }
                    if (intersectionIdx == (uint32_t)sourceCoords.size())
                    {
                        sourceCoords.push_back(sourceIntersection);
                        destCoords.push_back(intersection);
                    }
                    newFacet.m_vertices.push_back(intersectionIdx);
                    if (side < 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                    }

                    //iPrevIdx = iIdx;
                }
            }

            if (needFlip)
            {
                newFacets.emplace_back();
                auto& newFacet = newFacets.back();
                newFacet.m_trf = mirrorTrf * facet.m_trf;
                newFacet.m_invTrf = invert(newFacet.m_trf);
                newFacet.m_dir = (Dir)(1 - (uint32_t)facet.m_dir);

                uint32_t iPrevIdx = (uint32_t)facet.m_vertices.size() - 1;
                for (uint32_t iIdx = 0; iIdx < (uint32_t)facet.m_vertices.size(); iIdx++)
                {
                    uint32_t idx = facet.m_vertices[iIdx];
                    uint32_t prevIdx = facet.m_vertices[iPrevIdx];
                    auto& coord = destCoords[idx];
                    auto& prevCoord = destCoords[prevIdx];
                    int32_t side = sides[iIdx];
                    int32_t prevSide = sides[iPrevIdx];
                    iPrevIdx = iIdx;
                    if (side > 0 && prevSide >= 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side == 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side < 0 && prevSide <= 0)
                    {
                        continue;
                    }

                    mpq_class lineParam;
                    mpq_class segParam;
                    Coord segDir = coord - prevCoord;
                    if (!intersectLines(lineOrigin,
                                        lineDir,
                                        prevCoord,
                                        segDir,
                                        &lineParam,
                                        &segParam))
                    {
                        printf("Lines don't intersect???");
                        exit(1);
                    }

                    Coord intersection = prevCoord + segParam * segDir;
                    Coord sourceIntersection = facet.m_invTrf * intersection;
                    uint32_t intersectionIdx = 0;
                    for (; intersectionIdx < (uint32_t)sourceCoords.size(); intersectionIdx++)
                    {
                        if (sourceCoords[intersectionIdx] == sourceIntersection)
                        {
                            break;
                        }
                    }
                    if (intersectionIdx == (uint32_t)sourceCoords.size())
                    {
                        sourceCoords.push_back(sourceIntersection);
                        destCoords.push_back(intersection);
                    }
                    newFacet.m_vertices.push_back(intersectionIdx);
                    if (side > 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                    }

                    //iPrevIdx = iIdx;
                }
            }
        }

        facets = std::move(newFacets);

        for (uint32_t idx = 0; idx < (uint32_t)destCoords.size(); idx++)
        {
            Coord vec = destCoords[idx] - lineOrigin;
            int32_t side = sgn(cross(lineDir, vec));
            if (side > 0)
            {
                destCoords[idx] = mirrorTrf * destCoords[idx];
            }
        }
    };

    //connect(initTrf * Coord(0, 1), initTrf * Coord(mpq_class("1/2"), 0));

    /*
    connect(initTrf * Coord(mpq_class("0"), mpq_class("1")),
            initTrf * Coord(mpq_class("0"), mpq_class("3/4")));

    connect(initTrf * Coord(mpq_class("1"), mpq_class("7/8")),
            initTrf * Coord(mpq_class("1"), mpq_class("5/8")));

    connect(initTrf * Coord(mpq_class("1"), mpq_class("0")),
            initTrf * Coord(mpq_class("3/4"), mpq_class("0")));
    */

    auto getRandomCoord = [&](bool allowMidpoints) -> Coord
    {
        uint32_t iFacet = (uint32_t)(random.get() % facets.size());
        auto& facet = facets[iFacet];
        uint32_t iVertex = (uint32_t)(random.get() % facet.m_vertices.size());
        uint32_t vertexIdx = facet.m_vertices[iVertex];
        auto& coord = destCoords[vertexIdx];
        if (!allowMidpoints)
        {
            return coord;
        }

        uint32_t iNextVertex = (iVertex + 1) % (uint32_t)facet.m_vertices.size();
        uint32_t nextVertexIdx = facet.m_vertices[iNextVertex];
        auto& nextCoord = destCoords[nextVertexIdx];

        uint32_t factor = (uint32_t)(random.get() % 2 + 1);

        return nextCoord + (coord - nextCoord) / factor;
    };

    /*
    for (uint32_t iFold = 0; iFold < 6; iFold++)
    {
        Coord coord1;
        Coord coord2;
        do
        {
            coord1 = getRandomCoord(true);
            coord2 = getRandomCoord(false);
        }
        while (coord1 == coord2);
        connect(coord1, coord2);
    }
    */

    uint32_t folds = 0;
    while (true)
    {
        vector<Coord> sourceCoordsBak(sourceCoords);
        vector<Facet> facetsBak(facets);
        vector<Coord> destCoordsBak(destCoords);

        Coord coord1;
        Coord coord2;
        do
        {
            coord1 = getRandomCoord(true);
            coord2 = getRandomCoord(false);
        }
        while (coord1 == coord2);
        connect(coord1, coord2);

        solution.m_sourceCoords = sourceCoords;
        solution.m_destCoords = destCoords;
        solution.m_facets.clear();
        for (auto& facet : facets)
        {
            solution.m_facets.push_back(facet.m_vertices);
        }

        uint32_t solutionSize = 0;
        string msg;
        if (!formatSolution(solution, nullptr, &solutionSize, &msg))
        {
            printf("%s\n", msg.c_str());
            exit(1);
        }

        folds++;

        printf("%" PRIu32 " folds, solution size = %" PRIu32 "\n",
               folds,
               solutionSize);

        if (solutionSize > 1200)
        {
            sourceCoords = std::move(sourceCoordsBak);
            facets = std::move(facetsBak);
            destCoords = std::move(destCoordsBak);
            folds--;
            if (folds > 0)
            {
                break;
            }
        }
    }

    solution.m_sourceCoords = std::move(sourceCoords);
    solution.m_destCoords = std::move(destCoords);
    solution.m_facets.clear();
    for (auto& facet : facets)
    {
        solution.m_facets.push_back(std::move(facet.m_vertices));
    }
}
