#include "ConvexSolver.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Coord.hpp"
#include "Trf.hpp"
#include "Geometry.hpp"

using std::string;
using std::vector;

class Facet
{
public:
    Facet() : m_dir(Dir::CCW) {}

    vector<uint32_t> m_vertices;
    Trf m_trf;
    Trf m_invTrf;
    Dir m_dir;
};

bool solveConvex(const Problem& problem,
                 Solution* pSolution,
                 string* pSolutionData,
                 uint32_t* pSolutionSize)
{
    if (problem.m_polygons.size() != 1)
    {
        printf("# Not a single polygon\n");
        return false;
    }

    auto& polygon = problem.m_polygons[0];
    if (!isPolygonConvex(polygon))
    {
        printf("# Not convex\n");
        return false;
    }

    mpq_class xMin;
    mpq_class xMax;
    mpq_class yMin;
    mpq_class yMax;
    {
        bool first = true;
        for (auto& coord : polygon.m_vertices)
        {
            if (first)
            {
                first = false;
                xMin = coord.x;
                xMax = coord.x;
                yMin = coord.y;
                yMax = coord.y;
            }
            else
            {
                if (coord.x < xMin)
                {
                    xMin = coord.x;
                }
                if (coord.x > xMax)
                {
                    xMax = coord.x;
                }
                if (coord.y < yMin)
                {
                    yMin = coord.y;
                }
                if (coord.y > yMax)
                {
                    yMax = coord.y;
                }
            }
        }
    }

    if ((xMax - xMin > 1) ||
        (yMax - yMin > 1))
    {
        printf("# Does not fit in unrotated box\n");
        return false;
    }

    auto& solution = *pSolution;
    solution.m_sourceCoords.clear();
    solution.m_facets.clear();
    solution.m_destCoords.clear();

    vector<Coord> sourceCoords;
    sourceCoords.push_back(Coord(0, 0));
    sourceCoords.push_back(Coord(1, 0));
    sourceCoords.push_back(Coord(1, 1));
    sourceCoords.push_back(Coord(0, 1));

    vector<Coord> destCoords;

#if 1
    Trf initTrf = Trf(1, 0,
                      0, 1,
                      0, 0);
#endif

#if 0
    Trf initTrf = Trf(mpq_class("4/5"), mpq_class("-3/5"),
                      mpq_class("3/5"), mpq_class("4/5"),
                      0, 0);
#endif

#if 0
    initTrf.x += mpq_class("135246357468147258369470/13579086422468097531");
    initTrf.y += mpq_class("333774462228833923764587/66884455993322001177");
#endif

#if 0
    initTrf.x += mpq_class("2/5");
    initTrf.y += mpq_class("1/5");
#endif

#if 1
    initTrf.x += xMin;
    initTrf.y += yMin;
#endif

    Trf invInitTrf = invert(initTrf);

    vector<Facet> facets;
    {
        facets.emplace_back();
        auto& facet = facets.back();
        facet.m_vertices.push_back(0);
        facet.m_vertices.push_back(1);
        facet.m_vertices.push_back(2);
        facet.m_vertices.push_back(3);

        facet.m_dir = Dir::CCW;

        facet.m_trf = initTrf;
        facet.m_invTrf = invInitTrf;

        for (auto& sourceCoord : sourceCoords)
        {
            destCoords.push_back(facet.m_trf * sourceCoord);
        }
    }

    // Move coord1 to coord2
    auto connect = [&](const Coord& coord1, const Coord& coord2) -> bool
    {
        bool didAnything = false;

        Coord lineOrigin = (coord1 + coord2) / 2;
        Coord coord2dir = coord2 - lineOrigin;
        Coord lineDir(-coord2dir.y, coord2dir.x);

        Trf mirrorTrf = getMirrorTrf(lineOrigin, lineDir);

        /*
        printf("coord1 = (%s, %s)\n",
               coord1.x.get_str().c_str(),
               coord1.y.get_str().c_str());
        printf("coord2 = (%s, %s)\n",
               coord2.x.get_str().c_str(),
               coord2.y.get_str().c_str());
        printf("mirrorTrf(coord1) = (%s, %s)\n",
               (mirrorTrf*coord1).x.get_str().c_str(),
               (mirrorTrf*coord1).y.get_str().c_str());
        */

        vector<Facet> newFacets;

        uint32_t facetCount = (uint32_t)facets.size();
        for (uint32_t iFacet = 0; iFacet < facetCount; iFacet++)
        {
            bool needFlip = false;
            bool needOrig = false;
            auto& facet = facets[iFacet];
            vector<int32_t> sides;
            for (uint32_t vertexIdx : facet.m_vertices)
            {
                Coord vec = destCoords[vertexIdx] - lineOrigin;
                int32_t side = sgn(cross(lineDir, vec));
                sides.push_back(side);
                if (side > 0)
                {
                    needFlip = true;
                    didAnything = true;
                }
                else if (side < 0)
                {
                    needOrig = true;
                }
            }

            if (needOrig)
            {
                newFacets.emplace_back();
                auto& newFacet = newFacets.back();
                newFacet.m_trf = facet.m_trf;
                newFacet.m_invTrf = facet.m_invTrf;
                newFacet.m_dir = facet.m_dir;

                uint32_t iPrevIdx = (uint32_t)facet.m_vertices.size() - 1;
                for (uint32_t iIdx = 0; iIdx < (uint32_t)facet.m_vertices.size(); iIdx++)
                {
                    uint32_t idx = facet.m_vertices[iIdx];
                    uint32_t prevIdx = facet.m_vertices[iPrevIdx];
                    auto& coord = destCoords[idx];
                    auto& prevCoord = destCoords[prevIdx];
                    int32_t side = sides[iIdx];
                    int32_t prevSide = sides[iPrevIdx];
                    iPrevIdx = iIdx;
                    if (side < 0 && prevSide <= 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side == 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side > 0 && prevSide >= 0)
                    {
                        continue;
                    }

                    mpq_class lineParam;
                    mpq_class segParam;
                    Coord segDir = coord - prevCoord;
                    if (!intersectLines(lineOrigin,
                                        lineDir,
                                        prevCoord,
                                        segDir,
                                        &lineParam,
                                        &segParam))
                    {
                        printf("Lines don't intersect???");
                        exit(1);
                    }

                    Coord intersection = prevCoord + segParam * segDir;
                    Coord sourceIntersection = facet.m_invTrf * intersection;
                    uint32_t intersectionIdx = 0;
                    for (; intersectionIdx < (uint32_t)sourceCoords.size(); intersectionIdx++)
                    {
                        if (sourceCoords[intersectionIdx] == sourceIntersection)
                        {
                            break;
                        }
                    }
                    if (intersectionIdx == (uint32_t)sourceCoords.size())
                    {
                        sourceCoords.push_back(sourceIntersection);
                        destCoords.push_back(intersection);
                    }
                    newFacet.m_vertices.push_back(intersectionIdx);
                    if (side < 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                    }

                    //iPrevIdx = iIdx;
                }
            }

            if (needFlip)
            {
                newFacets.emplace_back();
                auto& newFacet = newFacets.back();
                newFacet.m_trf = mirrorTrf * facet.m_trf;
                newFacet.m_invTrf = invert(newFacet.m_trf);
                newFacet.m_dir = (Dir)(1 - (uint32_t)facet.m_dir);

                uint32_t iPrevIdx = (uint32_t)facet.m_vertices.size() - 1;
                for (uint32_t iIdx = 0; iIdx < (uint32_t)facet.m_vertices.size(); iIdx++)
                {
                    uint32_t idx = facet.m_vertices[iIdx];
                    uint32_t prevIdx = facet.m_vertices[iPrevIdx];
                    auto& coord = destCoords[idx];
                    auto& prevCoord = destCoords[prevIdx];
                    int32_t side = sides[iIdx];
                    int32_t prevSide = sides[iPrevIdx];
                    iPrevIdx = iIdx;
                    if (side > 0 && prevSide >= 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side == 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                        continue;
                    }
                    if (side < 0 && prevSide <= 0)
                    {
                        continue;
                    }

                    mpq_class lineParam;
                    mpq_class segParam;
                    Coord segDir = coord - prevCoord;
                    if (!intersectLines(lineOrigin,
                                        lineDir,
                                        prevCoord,
                                        segDir,
                                        &lineParam,
                                        &segParam))
                    {
                        printf("Lines don't intersect???");
                        exit(1);
                    }

                    Coord intersection = prevCoord + segParam * segDir;
                    Coord sourceIntersection = facet.m_invTrf * intersection;
                    uint32_t intersectionIdx = 0;
                    for (; intersectionIdx < (uint32_t)sourceCoords.size(); intersectionIdx++)
                    {
                        if (sourceCoords[intersectionIdx] == sourceIntersection)
                        {
                            break;
                        }
                    }
                    if (intersectionIdx == (uint32_t)sourceCoords.size())
                    {
                        sourceCoords.push_back(sourceIntersection);
                        destCoords.push_back(intersection);
                    }
                    newFacet.m_vertices.push_back(intersectionIdx);
                    if (side > 0)
                    {
                        newFacet.m_vertices.push_back(idx);
                    }

                    //iPrevIdx = iIdx;
                }
            }
        }

        facets = std::move(newFacets);

        for (uint32_t idx = 0; idx < (uint32_t)destCoords.size(); idx++)
        {
            Coord vec = destCoords[idx] - lineOrigin;
            int32_t side = sgn(cross(lineDir, vec));
            if (side > 0)
            {
                destCoords[idx] = mirrorTrf * destCoords[idx];
            }
        }

        return didAnything;
    };


    while (true)
    {
        bool didAnything = false;

        for (uint32_t idx = 0; idx < (uint32_t)polygon.m_vertices.size(); idx++)
        {
            auto& coord0 = polygon.m_vertices[idx];
            auto& coord1 = polygon.m_vertices[(idx + 1) % (uint32_t)polygon.m_vertices.size()];

            Coord vec = coord1 - coord0;

            Coord vecIn(-vec.y, vec.x);
            Coord vecOut(vec.y, -vec.x);
            Coord coordIn = coord0 + vecIn;
            Coord coordOut = coord0 + vecOut;

            if (connect(coordOut, coordIn))
            {
                didAnything = true;
            }
        }

        //printf("didAnything = %s\n", didAnything ? "true" : "false");

        if (!didAnything)
        {
            break;
        }
    }

    solution.m_sourceCoords = std::move(sourceCoords);
    solution.m_destCoords = std::move(destCoords);
    solution.m_facets.clear();
    for (auto& facet : facets)
    {
        solution.m_facets.push_back(std::move(facet.m_vertices));
    }

    uint32_t size = 0;
    string msg;
    string solutionData;
    if (!formatSolution(solution,
                        &solutionData,
                        &size,
                        &msg))
    {
        printf("formatSolution: %s\n", msg.c_str());
        return false;
    }

    if (pSolutionData)
    {
        *pSolutionData = std::move(solutionData);
    }
    if (pSolutionSize)
    {
        *pSolutionSize = size;
    }

    return true;
}
