#include "Problem.hpp"
#include "Cleanup.hpp"
#include "StringUtils.hpp"
#include <gmpxx.h>

using std::string;
using std::vector;

bool parseCount(vector<char>& data, size_t& pos, uint32_t* pCount)
{
    uint32_t count = 0;
    while (pos < data.size() && data[pos] >= '0' && data[pos] <= '9')
    {
        count *= 10;
        count += (data[pos] - '0');
        if (count > 5000)
        {
            return false;
        }
        pos++;
    }

    *pCount = count;
    return true;
}

bool parseNewline(vector<char>& data, size_t& pos)
{
    bool gotNewline = false;
    while (pos < data.size() && (data[pos] == '\n' || data[pos] == '\r'))
    {
        gotNewline = true;
        pos++;
    }

    return gotNewline;
}

bool parseUInt(vector<char>& data, size_t& pos)
{
    bool gotDigit = false;
    while (pos < data.size() && data[pos] >= '0' && data[pos] <= '9')
    {
        gotDigit = true;
        pos++;
    }
    return gotDigit;
}

bool parseRat(vector<char>& data, size_t& pos, mpq_class* pRat)
{
    size_t startPos = pos;
    if (pos < data.size() && data[pos] == '-')
    {
        pos++;
    }
    if (!parseUInt(data, pos))
    {
        return false;
    }
    if (pos < data.size() && data[pos] == '/')
    {
        pos++;
        if (!parseUInt(data, pos))
        {
            return false;
        }
    }

    string strRat(&data[startPos], pos - startPos);
    try
    {
        *pRat = strRat;
    }
    catch (const std::invalid_argument&)
    {
        return false;
    }

    pRat->canonicalize();

    return true;
}

bool parseCoord(vector<char>& data, size_t& pos, Coord* pCoord)
{
    auto& coord = *pCoord;
    if (!parseRat(data, pos, &coord.x))
    {
        return false;
    }
    if (pos < data.size() && data[pos] == ',')
    {
        pos++;
    }
    else
    {
        return false;
    }
    if (!parseRat(data, pos, &coord.y))
    {
        return false;
    }
    return true;
}

bool parseSegment(vector<char>& data, size_t& pos, Segment* pSegment)
{
    auto& segment = *pSegment;
    if (!parseCoord(data, pos, &segment.start))
    {
        return false;
    }
    if (pos < data.size() && data[pos] == ' ')
    {
        pos++;
    }
    else
    {
        return false;
    }
    if (!parseCoord(data, pos, &segment.end))
    {
        return false;
    }
    return true;
}

bool loadProblem(uint32_t id,
                 bool challengeProblem,
                 Problem* pProblem,
                 string* pMsg)
{
    string dummyMsg;
    if (!pMsg)
    {
        pMsg = &dummyMsg;
    }

    string fileName =
        challengeProblem ?
        strprintf("../challengeproblems/%" PRIu32 ".txt", id) :
        strprintf("../problems/%" PRIu32 ".txt", id);
    FILE* f = fopen(fileName.c_str(), "rb");
    if (!f)
    {
        *pMsg = "Error opening problem file";
        return false;
    }
    Cleanup cleanupFile([&](){ fclose(f); });

    fseek(f, 0, SEEK_END);
    size_t size = (size_t)ftell(f);
    fseek(f, 0, SEEK_SET);

    vector<char> data(size);
    if (fread(&data[0], 1, size, f) != size)
    {
        *pMsg = "Error reading problem file";
        return false;
    }
    cleanupFile.cleanup();

    auto& problem = *pProblem;

    size_t pos = 0;

    uint32_t polygonCount = 0;
    if (!parseCount(data, pos, &polygonCount))
    {
        *pMsg = "Error parsing polygon count";
        return false;
    }
    if (polygonCount < 1)
    {
        *pMsg = "Polygon count too low";
        return false;
    }

    problem.m_polygons.resize(polygonCount);

    for (uint32_t iPolygon = 0; iPolygon < polygonCount; iPolygon++)
    {
        auto& polygon = problem.m_polygons[iPolygon];

        if (!parseNewline(data, pos))
        {
            *pMsg = "Error parsing newline before vertex count";
            return false;
        }
        uint32_t vertexCount = 0;
        if (!parseCount(data, pos, &vertexCount))
        {
            *pMsg = "Error parsing vertex count";
            return false;
        }
        if (vertexCount < 3)
        {
            *pMsg = "Vertex count too low";
            return false;
        }

        polygon.m_vertices.resize(vertexCount);

        for (uint32_t iVertex = 0; iVertex < vertexCount; iVertex++)
        {
            auto& vertex = polygon.m_vertices[iVertex];

            if (!parseNewline(data, pos))
            {
                *pMsg = "Error parsing newline before vertex";
                return false;
            }
            if (!parseCoord(data, pos, &vertex))
            {
                *pMsg = "Error parsing vertex";
                return false;
            }
        }
    }

    if (!parseNewline(data, pos))
    {
        *pMsg = "Error parsing newline before segment count";
        return false;
    }
    uint32_t segmentCount = 0;
    if (!parseCount(data, pos, &segmentCount))
    {
        *pMsg = "Error parsing segment count";
        return false;
    }
    if (segmentCount < 3)
    {
        *pMsg = "Segment count too low";
        return false;
    }

    problem.m_segments.resize(segmentCount);

    for (uint32_t iSegment = 0; iSegment < segmentCount; iSegment++)
    {
        auto& segment = problem.m_segments[iSegment];

        if (!parseNewline(data, pos))
        {
            *pMsg = "Error parsing newline before segment";
            return false;
        }
        if (!parseSegment(data, pos, &segment))
        {
            *pMsg = "Error parsing segment";
            return false;
        }
    }

    parseNewline(data, pos);

    if (pos != data.size())
    {
        *pMsg = "Unexpected data at end of problem";
        return false;
    }

    return true;
}
