#include "Protocol.hpp"
#include "Cleanup.hpp"
#include "StringUtils.hpp"
#include "TimeUtils.hpp"
#include <curl/curl.h>
#include <jansson.h>

// sudo apt-get install libcurl4-openssl-dev
// sudo apt-get install libjansson-dev

using std::string;
using std::vector;
using std::pair;

namespace Protocol
{
    void init()
    {
        curl_global_init(CURL_GLOBAL_ALL);
    }

    void cleanup()
    {
        curl_global_cleanup();
    }

    size_t writeFunction(char* ptr, size_t size, size_t nmemb, void* userdata)
    {
        string* pStr = (string*)userdata;
        *pStr += string(ptr, size * nmemb);
        return size * nmemb;
    }

    bool getAPIKey(string* pKey)
    {
        FILE* f = fopen("../apikey.txt", "rb");
        if (!f)
        {
            return false;
        }

        char buf[100];
        size_t bytesRead = fread(buf, 1, 100, f);
        for (size_t i = 0; i < bytesRead; i++)
        {
            char ch = buf[i];
            if (ch < 33 || ch > 126)
            {
                break;
            }
            pKey->push_back(ch);
        }
        if (pKey->size() < 32)
        {
            return false;
        }

        return true;
    }
    
    bool makeRequest(const string& url,
                     const vector<pair<string, string>>* pPostData,
                     string* pResult,
                     string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string key;
        if (!getAPIKey(&key))
        {
            *pMsg = "Error reading API key";
            return false;
        }

        sleepMS(1500);

        string strResult;

        CURL* curl = curl_easy_init();
        Cleanup cleanupCurl([&](){ curl_easy_cleanup(curl); });

        struct curl_slist *headers = nullptr;
        headers = curl_slist_append(headers, "Expect:");
        string keyHeader = "X-API-Key: " + key;
        headers = curl_slist_append(headers, keyHeader.c_str());
        Cleanup cleanupHeaders([&](){ curl_slist_free_all(headers); });

        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFunction);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&strResult);
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
        curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "gzip");
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
        if (pPostData)
        {
            string postStr;
            bool first = true;
            for (auto& postItem : *pPostData)
            {
                if (first)
                {
                    first = false;
                }
                else
                {
                    postStr += "&";
                }

                {
                    char* key = curl_easy_escape(curl, postItem.first.c_str(), 0);
                    postStr += key;
                    curl_free(key);
                }

                postStr += "=";

                {
                    char* value = curl_easy_escape(curl, postItem.second.c_str(), 0);
                    postStr += value;
                    curl_free(value);
                }
            }

            curl_easy_setopt(curl, CURLOPT_POST, 1L);
            curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, postStr.c_str());
        }
        CURLcode res = curl_easy_perform(curl);
        if (res != CURLE_OK)
        {
            *pMsg = "curl request failed";
            return false;
        }

        long code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);

        //printf("%s\n", strResult.c_str());

        if (code != 200)
        {
            printf("%s\n", strResult.c_str());
            *pMsg = strprintf("curl response code %lu", code);
            return false;
        }

        if (pResult)
        {
            *pResult = std::move(strResult);
        }

        return true;
    }

    bool helloWorld(string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        if (!makeRequest("http://2016sv.icfpcontest.org/api/hello",
                         nullptr,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        return true;
    }

    bool getSnapshotList(vector<SnapshotRec>* pSnapshotRecs,
                         string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        if (!makeRequest("http://2016sv.icfpcontest.org/api/snapshot/list",
                         nullptr,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        auto& snapshotRecs = *pSnapshotRecs;

        json_error_t jerr;
        json_t* jSnapshotList = json_loads(strResult.c_str(), 0, &jerr);
        if (!jSnapshotList)
        {
            *pMsg = "Load snapshot list failed";
            return false;
        }
        Cleanup cleanupSnapshotList([&](){ json_decref(jSnapshotList); });
        if (!json_is_object(jSnapshotList))
        {
            *pMsg = "Snapshot list not object";
            return false;
        }

        json_t* jSnapshotRecs = json_object_get(jSnapshotList, "snapshots");
        if (!jSnapshotRecs)
        {
            *pMsg = "Error getting snapshot recs";
            return false;
        }
        if (!json_is_array(jSnapshotRecs))
        {
            *pMsg = "Snapshot recs not array";
            return false;
        }

        snapshotRecs.clear();

        uint32_t snapshotRecCount = (uint32_t)json_array_size(jSnapshotRecs);
        for (uint32_t iSnapshotRec = 0; iSnapshotRec < snapshotRecCount; iSnapshotRec++)
        {
            json_t* jSnapshotRec = json_array_get(jSnapshotRecs, iSnapshotRec);
            if (!jSnapshotRec)
            {
                *pMsg = "Error getting snapshot rec";
                return false;
            }
            if (!json_is_object(jSnapshotRec))
            {
                *pMsg = "Snapshot rec not object";
                return false;
            }

            snapshotRecs.emplace_back();
            auto& snapshotRec = snapshotRecs.back();

            {
                json_t* jTime = json_object_get(jSnapshotRec, "snapshot_time");
                if (!jTime)
                {
                    *pMsg = "Error getting snapshot rec time";
                    return false;
                }
                if (!json_is_integer(jTime))
                {
                    *pMsg = "Snapshot rec time not integer";
                    return false;
                }
                snapshotRec.m_time = (uint64_t)json_integer_value(jTime);
            }

            {
                json_t* jHash = json_object_get(jSnapshotRec, "snapshot_hash");
                if (!jHash)
                {
                    *pMsg = "Error getting snapshot rec hash";
                    return false;
                }
                if (!json_is_string(jHash))
                {
                    *pMsg = "Snapshot rec hash not string";
                    return false;
                }
                const char* rawHash = json_string_value(jHash);
                snapshotRec.m_hash = rawHash;
            }
        }

        return true;
    }

    bool getSnapshot(const string& snapshotHash,
                     Snapshot* pSnapshot,
                     string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        if (!makeRequest("http://2016sv.icfpcontest.org/api/blob/" +
                         snapshotHash,
                         nullptr,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        auto& snapshot = *pSnapshot;

        json_error_t jerr;
        json_t* jSnapshot = json_loads(strResult.c_str(), 0, &jerr);
        if (!jSnapshot)
        {
            *pMsg = "Load snapshot failed";
            return false;
        }
        Cleanup cleanupSnapshot([&](){ json_decref(jSnapshot); });
        if (!json_is_object(jSnapshot))
        {
            *pMsg = "Snapshot not object";
            return false;
        }

        {
            json_t* jProblemRecs = json_object_get(jSnapshot, "problems");
            if (!jProblemRecs)
            {
                *pMsg = "Error getting problem recs";
                return false;
            }
            if (!json_is_array(jProblemRecs))
            {
                *pMsg = "Problem recs not array";
                return false;
            }

            snapshot.m_problemRecs.clear();

            uint32_t problemRecCount = (uint32_t)json_array_size(jProblemRecs);
            for (uint32_t iProblemRec = 0; iProblemRec < problemRecCount; iProblemRec++)
            {
                json_t* jProblemRec = json_array_get(jProblemRecs, iProblemRec);
                if (!jProblemRec)
                {
                    *pMsg = "Error getting problem rec";
                    return false;
                }
                if (!json_is_object(jProblemRec))
                {
                    *pMsg = "Problem rec not object";
                    return false;
                }

                snapshot.m_problemRecs.emplace_back();
                auto& problemRec = snapshot.m_problemRecs.back();

                // Ranking
                {
                    json_t* jRankings = json_object_get(jProblemRec, "ranking");
                    if (!jRankings)
                    {
                        *pMsg = "Error getting problem rec rankings";
                        return false;
                    }
                    if (!json_is_array(jRankings))
                    {
                        *pMsg = "Problem rec rankings not array";
                        return false;
                    }

                    problemRec.m_rankings.clear();

                    uint32_t rankingCount = (uint32_t)json_array_size(jRankings);
                    for (uint32_t iRanking = 0; iRanking < rankingCount; iRanking++)
                    {
                        json_t* jRanking = json_array_get(jRankings, iRanking);
                        if (!jRanking)
                        {
                            *pMsg = "Error getting ranking";
                            return false;
                        }
                        if (!json_is_object(jRanking))
                        {
                            *pMsg = "Ranking not object";
                            return false;
                        }

                        problemRec.m_rankings.emplace_back();
                        auto& ranking = problemRec.m_rankings.back();

                        {
                            json_t* jResemblance = json_object_get(jRanking, "resemblance");
                            if (!jResemblance)
                            {
                                *pMsg = "Error getting ranking resemblance";
                                return false;
                            }
                            if (!json_is_real(jResemblance))
                            {
                                *pMsg = "Ranking resemblance not real";
                                return false;
                            }
                            ranking.m_resemblance = json_real_value(jResemblance);
                        }

                        {
                            json_t* jSolutionSize = json_object_get(jRanking, "solution_size");
                            if (!jSolutionSize)
                            {
                                *pMsg = "Error getting ranking solution size";
                                return false;
                            }
                            if (!json_is_integer(jSolutionSize))
                            {
                                *pMsg = "Ranking solution size not integer";
                                return false;
                            }
                            ranking.m_solutionSize = (uint32_t)json_integer_value(jSolutionSize);
                        }
                    }
                }

                {
                    json_t* jTime = json_object_get(jProblemRec, "publish_time");
                    if (!jTime)
                    {
                        *pMsg = "Error getting problem rec time";
                        return false;
                    }
                    if (!json_is_integer(jTime))
                    {
                        *pMsg = "Problem rec time not integer";
                        return false;
                    }
                    problemRec.m_time = (uint64_t)json_integer_value(jTime);
                }

                {
                    json_t* jSolutionSize = json_object_get(jProblemRec, "solution_size");
                    if (!jSolutionSize)
                    {
                        *pMsg = "Error getting problem rec solution size";
                        return false;
                    }
                    if (!json_is_integer(jSolutionSize))
                    {
                        *pMsg = "Problem rec solution size not integer";
                        return false;
                    }
                    problemRec.m_solutionSize = (uint32_t)json_integer_value(jSolutionSize);
                }

                {
                    json_t* jId = json_object_get(jProblemRec, "problem_id");
                    if (!jId)
                    {
                        *pMsg = "Error getting problem rec ID";
                        return false;
                    }
                    if (!json_is_integer(jId))
                    {
                        *pMsg = "Problem rec ID not integer";
                        return false;
                    }
                    problemRec.m_id = (uint32_t)json_integer_value(jId);
                }

                {
                    json_t* jOwner = json_object_get(jProblemRec, "owner");
                    if (!jOwner)
                    {
                        *pMsg = "Error getting problem rec owner";
                        return false;
                    }
                    if (!json_is_string(jOwner))
                    {
                        *pMsg = "Problem rec owner not string";
                        return false;
                    }
                    const char* rawOwner = json_string_value(jOwner);
                    problemRec.m_owner = rawOwner;
                }

                {
                    json_t* jProblemSize = json_object_get(jProblemRec, "problem_size");
                    if (!jProblemSize)
                    {
                        *pMsg = "Error getting problem rec problem size";
                        return false;
                    }
                    if (!json_is_integer(jProblemSize))
                    {
                        *pMsg = "Problem rec problem size not integer";
                        return false;
                    }
                    problemRec.m_problemSize = (uint32_t)json_integer_value(jProblemSize);
                }

                {
                    json_t* jHash = json_object_get(jProblemRec, "problem_spec_hash");
                    if (!jHash)
                    {
                        *pMsg = "Error getting problem rec hash";
                        return false;
                    }
                    if (!json_is_string(jHash))
                    {
                        *pMsg = "Problem rec hash not string";
                        return false;
                    }
                    const char* rawHash = json_string_value(jHash);
                    problemRec.m_hash = rawHash;
                }
            }
        }

        {
            json_t* jTime = json_object_get(jSnapshot, "snapshot_time");
            if (!jTime)
            {
                *pMsg = "Error getting snapshot time";
                return false;
            }
            if (!json_is_integer(jTime))
            {
                *pMsg = "Snapshot time not integer";
                return false;
            }
            snapshot.m_time = (uint64_t)json_integer_value(jTime);
        }

        {
            json_t* jLeaderboardRecs = json_object_get(jSnapshot, "leaderboard");
            if (!jLeaderboardRecs)
            {
                *pMsg = "Error getting leaderboard recs";
                return false;
            }
            if (!json_is_array(jLeaderboardRecs))
            {
                *pMsg = "Leaderboard recs not array";
                return false;
            }

            snapshot.m_leaderboardRecs.clear();

            uint32_t leaderboardRecCount = (uint32_t)json_array_size(jLeaderboardRecs);
            for (uint32_t iLeaderboardRec = 0; iLeaderboardRec < leaderboardRecCount; iLeaderboardRec++)
            {
                json_t* jLeaderboardRec = json_array_get(jLeaderboardRecs, iLeaderboardRec);
                if (!jLeaderboardRec)
                {
                    *pMsg = "Error getting leaderboard rec";
                    return false;
                }
                if (!json_is_object(jLeaderboardRec))
                {
                    *pMsg = "Leaderboard rec not object";
                    return false;
                }

                snapshot.m_leaderboardRecs.emplace_back();
                auto& leaderboardRec = snapshot.m_leaderboardRecs.back();

                {
                    json_t* jUsername = json_object_get(jLeaderboardRec, "username");
                    if (!jUsername)
                    {
                        *pMsg = "Error getting leaderboard rec username";
                        return false;
                    }
                    if (!json_is_string(jUsername))
                    {
                        *pMsg = "Leaderboard rec username not string";
                        return false;
                    }
                    const char* rawUsername = json_string_value(jUsername);
                    leaderboardRec.m_username = rawUsername;
                }

                {
                    json_t* jScore = json_object_get(jLeaderboardRec, "score");
                    if (!jScore)
                    {
                        *pMsg = "Error getting leaderboard score";
                        return false;
                    }
                    if (!json_is_real(jScore))
                    {
                        *pMsg = "Leaderboard rec score not real";
                        return false;
                    }
                    leaderboardRec.m_score = json_real_value(jScore);
                }
            }
        }

        {
            json_t* jUsers = json_object_get(jSnapshot, "users");
            if (!jUsers)
            {
                *pMsg = "Error getting users";
                return false;
            }
            if (!json_is_array(jUsers))
            {
                *pMsg = "Users not array";
                return false;
            }

            snapshot.m_users.clear();

            uint32_t userCount = (uint32_t)json_array_size(jUsers);
            for (uint32_t iUser = 0; iUser < userCount; iUser++)
            {
                json_t* jUser = json_array_get(jUsers, iUser);
                if (!jUser)
                {
                    *pMsg = "Error getting user";
                    return false;
                }
                if (!json_is_object(jUser))
                {
                    *pMsg = "User not object";
                    return false;
                }

                snapshot.m_users.emplace_back();
                auto& user = snapshot.m_users.back();

                {
                    json_t* jUsername = json_object_get(jUser, "username");
                    if (!jUsername)
                    {
                        *pMsg = "Error getting user username";
                        return false;
                    }
                    if (!json_is_string(jUsername))
                    {
                        *pMsg = "User username not string";
                        return false;
                    }
                    const char* rawUsername = json_string_value(jUsername);
                    user.m_username = rawUsername;
                }

                {
                    json_t* jDisplayName = json_object_get(jUser, "display_name");
                    if (!jDisplayName)
                    {
                        *pMsg = "Error getting user display name";
                        return false;
                    }
                    if (!json_is_string(jDisplayName))
                    {
                        *pMsg = "User display name not string";
                        return false;
                    }
                    const char* rawDisplayName = json_string_value(jDisplayName);
                    user.m_displayName = rawDisplayName;
                }
            }
        }

        return true;
    }

    bool getProblem(const string& problemHash,
                    string* pProblem,
                    string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        if (!makeRequest("http://2016sv.icfpcontest.org/api/blob/" +
                         problemHash,
                         nullptr,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        *pProblem = std::move(strResult);

        return true;
    }

    bool submitSolution(uint32_t id,
                        const string& solution,
                        SolutionResult* pResult,
                        string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        vector<pair<string, string>> postData;
        postData.emplace_back("problem_id", strprintf("%" PRIu32 "", id));
        postData.emplace_back("solution_spec", solution);

        if (!makeRequest("http://2016sv.icfpcontest.org/api/solution/submit",
                         &postData,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        auto& result = *pResult;

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed";
            return false;
        }
        Cleanup cleanupResult([&](){ json_decref(jResult); });
        if (!json_is_object(jResult))
        {
            *pMsg = "Result not object";
            return false;
        }

        {
            json_t* jId = json_object_get(jResult, "problem_id");
            if (!jId)
            {
                *pMsg = "Error getting problem ID";
                return false;
            }
            if (!json_is_integer(jId))
            {
                *pMsg = "Problem ID not integer";
                return false;
            }
            result.m_id = (uint32_t)json_integer_value(jId);
        }

        {
            json_t* jResemblance = json_object_get(jResult, "resemblance");
            if (!jResemblance)
            {
                *pMsg = "Error getting resemblance";
                return false;
            }
            if (!json_is_real(jResemblance))
            {
                *pMsg = "Resemblance not real";
                return false;
            }
            result.m_resemblance = json_real_value(jResemblance);
        }

        {
            json_t* jHash = json_object_get(jResult, "solution_spec_hash");
            if (!jHash)
            {
                *pMsg = "Error getting solution hash";
                return false;
            }
            if (!json_is_string(jHash))
            {
                *pMsg = "Solution hash not string";
                return false;
            }
            const char* rawHash = json_string_value(jHash);
            result.m_solutionHash = rawHash;
        }

        {
            json_t* jSize = json_object_get(jResult, "solution_size");
            if (!jSize)
            {
                *pMsg = "Error getting solution size";
                return false;
            }
            if (!json_is_integer(jSize))
            {
                *pMsg = "Solution size not integer";
                return false;
            }
            result.m_solutionSize = (uint32_t)json_integer_value(jSize);
        }

        return true;
    }

    bool submitProblem(uint64_t time,
                       const string& solution,
                       ProblemResult* pResult,
                       string* pMsg)
    {
        string dummyMsg;
        if (!pMsg)
        {
            pMsg = &dummyMsg;
        }

        string strResult;

        vector<pair<string, string>> postData;
        postData.emplace_back("solution_spec", solution);
        postData.emplace_back("publish_time", strprintf("%" PRIu64 "", time));

        if (!makeRequest("http://2016sv.icfpcontest.org/api/problem/submit",
                         &postData,
                         &strResult,
                         pMsg))
        {
            return false;
        }

        auto& result = *pResult;

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed";
            return false;
        }
        Cleanup cleanupResult([&](){ json_decref(jResult); });
        if (!json_is_object(jResult))
        {
            *pMsg = "Result not object";
            return false;
        }

        {
            json_t* jId = json_object_get(jResult, "problem_id");
            if (!jId)
            {
                *pMsg = "Error getting problem ID";
                return false;
            }
            if (!json_is_integer(jId))
            {
                *pMsg = "Problem ID not integer";
                return false;
            }
            result.m_id = (uint32_t)json_integer_value(jId);
        }

        {
            json_t* jTime = json_object_get(jResult, "publish_time");
            if (!jTime)
            {
                *pMsg = "Error getting publish time";
                return false;
            }
            if (!json_is_integer(jTime))
            {
                *pMsg = "Publish time not integer";
                return false;
            }
            result.m_time = (uint32_t)json_integer_value(jTime);
        }

        {
            json_t* jHash = json_object_get(jResult, "solution_spec_hash");
            if (!jHash)
            {
                *pMsg = "Error getting solution hash";
                return false;
            }
            if (!json_is_string(jHash))
            {
                *pMsg = "Solution hash not string";
                return false;
            }
            const char* rawHash = json_string_value(jHash);
            result.m_solutionHash = rawHash;
        }

        {
            json_t* jSize = json_object_get(jResult, "solution_size");
            if (!jSize)
            {
                *pMsg = "Error getting solution size";
                return false;
            }
            if (!json_is_integer(jSize))
            {
                *pMsg = "Solution size not integer";
                return false;
            }
            result.m_solutionSize = (uint32_t)json_integer_value(jSize);
        }

        {
            json_t* jHash = json_object_get(jResult, "problem_spec_hash");
            if (!jHash)
            {
                *pMsg = "Error getting problem hash";
                return false;
            }
            if (!json_is_string(jHash))
            {
                *pMsg = "Problem hash not string";
                return false;
            }
            const char* rawHash = json_string_value(jHash);
            result.m_problemHash = rawHash;
        }

        {
            json_t* jSize = json_object_get(jResult, "problem_size");
            if (!jSize)
            {
                *pMsg = "Error getting problem size";
                return false;
            }
            if (!json_is_integer(jSize))
            {
                *pMsg = "Problem size not integer";
                return false;
            }
            result.m_problemSize = (uint32_t)json_integer_value(jSize);
        }

        return true;
    }
}
