#include "Solver.hpp"
#include "Problem.hpp"
#include "Coord.hpp"
#include "Geometry.hpp"
#include "Trf.hpp"
#include "Solution.hpp"
#include "TimeUtils.hpp"
#include "Cleanup.hpp"
#include "Random.hpp"

using std::string;
using std::vector;

uint32_t Shape::findVertex(const Coord& vertex) const
{
    ShapeVertex shapeVertex;
    shapeVertex.m_coord = vertex;

    auto it = std::lower_bound(m_vertices.begin(),
                               m_vertices.end(),
                               shapeVertex,
                               [](const ShapeVertex& sv1,
                                  const ShapeVertex& sv2)
                               { return sv1.m_coord < sv2.m_coord; });
    if (it == m_vertices.end() ||
        shapeVertex.m_coord < it->m_coord)
    {
        return (uint32_t)-1;
    }

    return (uint32_t)(it - m_vertices.begin());
}

void splitSegments(const vector<Segment>& segments,
                   vector<Segment>* pSplitSegments)
{
    pSplitSegments->clear();
    pSplitSegments->reserve(segments.size());

    for (uint32_t i = 0; i < (uint32_t)segments.size(); i++)
    {
        auto& s1 = segments[i];
        vector<Coord> coords;
        coords.reserve(100);
        coords.push_back(s1.start);
        coords.push_back(s1.end);

        for (uint32_t j = 0; j < (uint32_t)segments.size(); j++)
        {
            if (j == i)
            {
                continue;
            }

            auto& s2 = segments[j];

            Coord d1 = s1.end - s1.start;
            Coord d2 = s2.end - s2.start;
            mpq_class dx = cross(d1, d2);
            if (dx == 0)
            {
                // Segments are Parallel
                continue;
            }
            Coord ds = s2.start - s1.start;
            mpq_class t = cross(ds, d2) / dx;
            mpq_class u = cross(ds, d1) / dx;

            if (t <= 0 || t >= 1 || u < 0 || u > 1)
            {
                // Segments do not intersect the interior of s1
                continue;
            }

            if (u == 0)
            {
                coords.push_back(s2.start);
            }
            else if (u == 1)
            {
                coords.push_back(s2.end);
            }
            else
            {
                coords.push_back(s2.start + u * d2);
            }
        }

        std::sort(coords.begin(), coords.end());
        coords.erase(std::unique(coords.begin(), coords.end()), coords.end());

        for (uint32_t j = 1; j < (uint32_t)coords.size(); j++)
        {
            pSplitSegments->emplace_back(coords[j - 1], coords[j]);
        }
    }
}

void problemToShape(const Problem& problem,
                    Shape* pShape)
{
    //Shape shape;
    auto& shape = *pShape;
    shape.m_vertices.clear();
    shape.m_edges.clear();
    shape.m_regions.clear();

    {
        vector<Segment> segments;
        splitSegments(problem.m_segments, &segments);

        vector<Coord> vertices;
        vertices.reserve(segments.size() * 2);
        for (auto& segment : segments)
        {
            vertices.push_back(segment.start);
            vertices.push_back(segment.end);
        }

        std::sort(vertices.begin(), vertices.end());
        vertices.erase(std::unique(vertices.begin(), vertices.end()),
                       vertices.end());

        shape.m_vertices.resize(vertices.size());
        for (uint32_t i = 0; i < (uint32_t)vertices.size(); i++)
        {
            shape.m_vertices[i].m_coord = std::move(vertices[i]);
        }

        shape.m_edges.resize(segments.size());
        for (uint32_t i = 0; i < (uint32_t)segments.size(); i++)
        {
            auto& edge = shape.m_edges[i];
            auto& segment = segments[i];
            edge.m_vertices[0].m_vertexId = shape.findVertex(segment.start);
            edge.m_vertices[1].m_vertexId = shape.findVertex(segment.end);
        }
    }

    for (uint32_t edgeId = 0; edgeId < (uint32_t)shape.m_edges.size(); edgeId++)
    {
        auto& edge = shape.m_edges[edgeId];
        for (uint32_t vertexIdx = 0; vertexIdx < 2; vertexIdx++)
        {
            uint32_t vertexId = edge.m_vertices[vertexIdx].m_vertexId;
            auto& vertex = shape.m_vertices[vertexId];
            vertex.m_edges.emplace_back();
            auto& vertexEdge = vertex.m_edges.back();
            vertexEdge.m_edgeId = edgeId;
            vertexEdge.m_vertexIdx = vertexIdx;
        }
    }

    for (uint32_t vertexId = 0; vertexId < (uint32_t)shape.m_vertices.size(); vertexId++)
    {
        auto& vertex = shape.m_vertices[vertexId];

        auto comp = [&](const ShapeVertexEdge& ve1,
                        const ShapeVertexEdge& ve2) -> bool
        {
            Coord vec1 = shape.m_vertices[shape.m_edges[ve1.m_edgeId].m_vertices[1 - ve1.m_vertexIdx].m_vertexId].m_coord - vertex.m_coord;
            Coord vec2 = shape.m_vertices[shape.m_edges[ve2.m_edgeId].m_vertices[1 - ve2.m_vertexIdx].m_vertexId].m_coord - vertex.m_coord;

            if (vec2.y == 0 && vec2.x > 0)
            {
                return false;
            }

            if (vec1.y == 0 && vec1.x > 0)
            {
                return true;
            }

            if (vec1.y == 0)
            {
                return vec2.y < 0;
            }

            if (vec2.y == 0)
            {
                return vec1.y > 0;
            }

            if (vec1.y > 0)
            {
                if (vec2.y < 0)
                {
                    return true;
                }

                //return (vec1.x / vec1.y) > (vec2.x / vec2.y);
                return (vec1.x * vec2.y) > (vec2.x * vec1.y);
            }

            if (vec2.y > 0)
            {
                return false;
            }

            //return (vec1.x / vec1.y) > (vec2.x / vec2.y);
            return (vec1.x * vec2.y) > (vec2.x * vec1.y);
        };

        std::sort(vertex.m_edges.begin(), vertex.m_edges.end(), comp);

        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
        {
            auto& ve = vertex.m_edges[edgeIdx];
            shape.m_edges[ve.m_edgeId].m_vertices[ve.m_vertexIdx].m_edgeIdx = edgeIdx;
        }
    }

    for (uint32_t edgeId = 0; edgeId < (uint32_t)shape.m_edges.size(); edgeId++)
    {
        auto& edge = shape.m_edges[edgeId];
        for (uint32_t regionIdx = 0; regionIdx < 2; regionIdx++)
        {
            auto& er = edge.m_regions[regionIdx];
            if (er.m_regionId != (uint32_t)-1)
            {
                continue;
            }

            //mpq_class sum(0);

            uint32_t regionId = (uint32_t)shape.m_regions.size();
            shape.m_regions.emplace_back();
            auto& region = shape.m_regions.back();
            region.m_type = ShapeRegionType::Interior;

            {
                auto& edgeStart = shape.m_vertices[edge.m_vertices[regionIdx].m_vertexId].m_coord;
                auto& edgeEnd = shape.m_vertices[edge.m_vertices[1 - regionIdx].m_vertexId].m_coord;
                for (auto& polygon : problem.m_polygons)
                {
                    const Coord* pPrevCoord = &polygon.m_vertices.back();
                    for (auto& coord : polygon.m_vertices)
                    {
                        auto& prevCoord = *pPrevCoord;
                        pPrevCoord = &coord;
                        Coord polyEdgeDir = coord - prevCoord;
                        if (cross(polyEdgeDir, edgeStart - prevCoord) != 0 ||
                            cross(polyEdgeDir, edgeEnd - prevCoord) != 0)
                        {
                            continue;
                        }

                        Coord projStart1 = project(edgeStart - prevCoord,
                                                   polyEdgeDir);
                        Coord projEnd1 = project(edgeEnd - prevCoord,
                                                 polyEdgeDir);

                        mpq_class ds1 = dot(projStart1, polyEdgeDir);
                        mpq_class de1 = dot(projEnd1, polyEdgeDir);

                        if (ds1 < 0 || de1 < 0)
                        {
                            continue;
                        }

                        Coord polyEdgeDir2 = prevCoord - coord;

                        Coord projStart2 = project(edgeStart - coord,
                                                   polyEdgeDir2);
                        Coord projEnd2 = project(edgeEnd - coord,
                                                 polyEdgeDir2);

                        mpq_class ds2 = dot(projStart2, polyEdgeDir2);
                        mpq_class de2 = dot(projEnd2, polyEdgeDir2);

                        if (ds2 < 0 || de2 < 0)
                        {
                            continue;
                        }

                        if (ds2 < de2)
                        {
                            region.m_type = ShapeRegionType::Exterior;
                        }
                    }
                }
            }

            uint32_t curEdgeId = edgeId;
            uint32_t curRegionIdx = regionIdx;
            do
            {
                uint32_t curEdgeIdx = (uint32_t)region.m_edges.size();
                region.m_edges.emplace_back();
                auto& curRe = region.m_edges.back();
                curRe.m_edgeId = curEdgeId;
                curRe.m_regionIdx = curRegionIdx;
                auto& curEdge = shape.m_edges[curEdgeId];
                auto& curEr = curEdge.m_regions[curRegionIdx];
                curEr.m_regionId = regionId;
                curEr.m_edgeIdx = curEdgeIdx;

                //auto& prevEv = curEdge.m_vertices[curRegionIdx];
                //uint32_t prevVertexId = prevEv.m_vertexId;
                //auto& prevVertex = shape.m_vertices[prevVertexId];

                auto& curEv = curEdge.m_vertices[1 - curRegionIdx];
                uint32_t curVertexId = curEv.m_vertexId;
                auto& curVertex = shape.m_vertices[curVertexId];

                //sum += prevVertex.m_coord.x * curVertex.m_coord.y;
                //sum -= prevVertex.m_coord.y * curVertex.m_coord.x;

                auto& nextVe = curVertex.m_edges[(curEv.m_edgeIdx + (uint32_t)curVertex.m_edges.size() - 1) % (uint32_t)curVertex.m_edges.size()];

                curEdgeId = nextVe.m_edgeId;
                curRegionIdx = nextVe.m_vertexIdx;
            }
            while (curEdgeId != edgeId);

            /*
            if (sum > 0)
            {
                region.m_type = ShapeRegionType::Interior;
            }
            else
            {
                region.m_type = ShapeRegionType::Exterior;
            }
            */
        }
    }
}

class SolverDoneException {};
class SolverTimeoutException {};
class SolverErrorException {};

class Solver
{
public:
    Solver(uint64_t seed = 1) :
        m_nodeCount(0),
        m_stopTimeMS(0),
        m_solved(false),
        m_solutionSize(0),
        m_random(seed)
    {
    }

    void solve(const Problem& problem,
               uint32_t maxDurationMS = 0);
    bool addRegion(uint32_t destRegionId,
                   uint32_t sourceRegionId,
                   const Trf& trf,
                   const Trf& invTrf,
                   uint32_t flipped);
    void recurse();
    void prepareSolution();

    Shape destShape;
    Shape sourceShape;
    vector<uint32_t> destUseCounts;

    uint64_t m_nodeCount;
    uint64_t m_stopTimeMS;

    bool m_solved;
    Solution m_solution;
    string m_solutionData;
    uint32_t m_solutionSize;

    Random m_random;
};

void Solver::solve(const Problem& problem, uint32_t maxDurationMS)
{
    if (maxDurationMS != 0)
    {
        m_stopTimeMS = getTimeMS() + maxDurationMS;
    }
    m_nodeCount = 0;

    problemToShape(problem, &destShape);

    destUseCounts.resize(destShape.m_regions.size());

    sourceShape.m_vertices.resize(4);
    sourceShape.m_vertices[0].m_coord = Coord(0, 0);
    sourceShape.m_vertices[1].m_coord = Coord(1, 0);
    sourceShape.m_vertices[2].m_coord = Coord(1, 1);
    sourceShape.m_vertices[3].m_coord = Coord(0, 1);

    sourceShape.m_edges.resize(4);
    sourceShape.m_edges[0].m_vertices[0].m_vertexId = 0;
    sourceShape.m_edges[0].m_vertices[1].m_vertexId = 1;
    sourceShape.m_edges[1].m_vertices[0].m_vertexId = 1;
    sourceShape.m_edges[1].m_vertices[1].m_vertexId = 2;
    sourceShape.m_edges[2].m_vertices[0].m_vertexId = 2;
    sourceShape.m_edges[2].m_vertices[1].m_vertexId = 3;
    sourceShape.m_edges[3].m_vertices[0].m_vertexId = 3;
    sourceShape.m_edges[3].m_vertices[1].m_vertexId = 0;

    sourceShape.m_vertices[0].m_edges.resize(2);
    sourceShape.m_vertices[0].m_edges[0].m_edgeId = 3;
    sourceShape.m_vertices[0].m_edges[0].m_vertexIdx = 1;
    sourceShape.m_edges[3].m_vertices[1].m_edgeIdx = 0;
    sourceShape.m_vertices[0].m_edges[1].m_edgeId = 0;
    sourceShape.m_vertices[0].m_edges[1].m_vertexIdx = 0;
    sourceShape.m_edges[0].m_vertices[0].m_edgeIdx = 1;

    sourceShape.m_vertices[1].m_edges.resize(2);
    sourceShape.m_vertices[1].m_edges[0].m_edgeId = 0;
    sourceShape.m_vertices[1].m_edges[0].m_vertexIdx = 1;
    sourceShape.m_edges[0].m_vertices[1].m_edgeIdx = 0;
    sourceShape.m_vertices[1].m_edges[1].m_edgeId = 1;
    sourceShape.m_vertices[1].m_edges[1].m_vertexIdx = 0;
    sourceShape.m_edges[1].m_vertices[0].m_edgeIdx = 1;

    sourceShape.m_vertices[2].m_edges.resize(2);
    sourceShape.m_vertices[2].m_edges[0].m_edgeId = 1;
    sourceShape.m_vertices[2].m_edges[0].m_vertexIdx = 1;
    sourceShape.m_edges[1].m_vertices[1].m_edgeIdx = 0;
    sourceShape.m_vertices[2].m_edges[1].m_edgeId = 2;
    sourceShape.m_vertices[2].m_edges[1].m_vertexIdx = 0;
    sourceShape.m_edges[2].m_vertices[0].m_edgeIdx = 1;

    sourceShape.m_vertices[3].m_edges.resize(2);
    sourceShape.m_vertices[3].m_edges[0].m_edgeId = 2;
    sourceShape.m_vertices[3].m_edges[0].m_vertexIdx = 1;
    sourceShape.m_edges[2].m_vertices[1].m_edgeIdx = 0;
    sourceShape.m_vertices[3].m_edges[1].m_edgeId = 3;
    sourceShape.m_vertices[3].m_edges[1].m_vertexIdx = 0;
    sourceShape.m_edges[3].m_vertices[0].m_edgeIdx = 1;

    sourceShape.m_regions.resize(2);

    sourceShape.m_regions[0].m_type = ShapeRegionType::Interior;
    sourceShape.m_regions[0].m_flipped = 0;
    sourceShape.m_regions[0].m_used = 0;
    sourceShape.m_regions[0].m_edges.resize(4);

    sourceShape.m_regions[0].m_edges[0].m_edgeId = 0;
    sourceShape.m_regions[0].m_edges[0].m_regionIdx = 0;
    sourceShape.m_edges[0].m_regions[0].m_regionId = 0;
    sourceShape.m_edges[0].m_regions[0].m_edgeIdx = 0;

    sourceShape.m_regions[0].m_edges[1].m_edgeId = 1;
    sourceShape.m_regions[0].m_edges[1].m_regionIdx = 0;
    sourceShape.m_edges[1].m_regions[0].m_regionId = 0;
    sourceShape.m_edges[1].m_regions[0].m_edgeIdx = 1;

    sourceShape.m_regions[0].m_edges[2].m_edgeId = 2;
    sourceShape.m_regions[0].m_edges[2].m_regionIdx = 0;
    sourceShape.m_edges[2].m_regions[0].m_regionId = 0;
    sourceShape.m_edges[2].m_regions[0].m_edgeIdx = 2;

    sourceShape.m_regions[0].m_edges[3].m_edgeId = 3;
    sourceShape.m_regions[0].m_edges[3].m_regionIdx = 0;
    sourceShape.m_edges[3].m_regions[0].m_regionId = 0;
    sourceShape.m_edges[3].m_regions[0].m_edgeIdx = 3;

    sourceShape.m_regions[1].m_type = ShapeRegionType::Exterior;
    sourceShape.m_regions[1].m_flipped = 0;
    sourceShape.m_regions[1].m_edges.resize(4);

    sourceShape.m_regions[1].m_edges[0].m_edgeId = 3;
    sourceShape.m_regions[1].m_edges[0].m_regionIdx = 1;
    sourceShape.m_edges[3].m_regions[1].m_regionId = 1;
    sourceShape.m_edges[3].m_regions[1].m_edgeIdx = 0;

    sourceShape.m_regions[1].m_edges[1].m_edgeId = 2;
    sourceShape.m_regions[1].m_edges[1].m_regionIdx = 1;
    sourceShape.m_edges[2].m_regions[1].m_regionId = 1;
    sourceShape.m_edges[2].m_regions[1].m_edgeIdx = 1;

    sourceShape.m_regions[1].m_edges[2].m_edgeId = 1;
    sourceShape.m_regions[1].m_edges[2].m_regionIdx = 1;
    sourceShape.m_edges[1].m_regions[1].m_regionId = 1;
    sourceShape.m_edges[1].m_regions[1].m_edgeIdx = 2;

    sourceShape.m_regions[1].m_edges[3].m_edgeId = 0;
    sourceShape.m_regions[1].m_edges[3].m_regionIdx = 1;
    sourceShape.m_edges[0].m_regions[1].m_regionId = 1;
    sourceShape.m_edges[0].m_regions[1].m_edgeIdx = 3;

    /*
    vector<Coord> sourceCoords;
    sourceCoords.push_back(Coord(0, 0));
    sourceCoords.push_back(Coord(1, 0));
    sourceCoords.push_back(Coord(1, 1));
    sourceCoords.push_back(Coord(0, 1));

    {
        facets.emplace_back();
        auto& facet = facets.back();
        facet.m_vertices.push_back(0);
        facet.m_vertices.push_back(1);
        facet.m_vertices.push_back(2);
        facet.m_vertices.push_back(3);
        facet.m_used = false;
    }
    */

    // Find first facet

    vector<uint32_t> regionIds(destShape.m_regions.size());
    for (uint32_t regionId = 0; regionId < (uint32_t)destShape.m_regions.size(); regionId++)
    {
        regionIds[regionId] = regionId;
    }
#if 1
    {
        uint32_t idx = (uint32_t)destShape.m_regions.size() - 1;
        while (idx > 0)
        {
            using std::swap;
            swap(regionIds[idx], regionIds[m_random.get() % (idx + 1)]);
            idx--;
        }
    }
#endif

    //for (uint32_t regionId = 0; regionId < (uint32_t)destShape.m_regions.size(); regionId++)
    for (uint32_t regionId : regionIds)
    {
        auto& region = destShape.m_regions[regionId];
        if (region.m_type == ShapeRegionType::Exterior)
        {
            continue;
        }
        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)region.m_edges.size(); edgeIdx++)
        {
            auto& re = region.m_edges[edgeIdx];
            uint32_t edgeId = re.m_edgeId;
            uint32_t regionIdx = re.m_regionIdx;
            auto& edge = destShape.m_edges[edgeId];

            auto& coord0 = destShape.m_vertices[edge.m_vertices[regionIdx].m_vertexId].m_coord;
            auto& coord1 = destShape.m_vertices[edge.m_vertices[1 - regionIdx].m_vertexId].m_coord;
            Coord vec = coord1 - coord0;

            mpz_class sq =
                (vec.x.get_num() * vec.x.get_num() *
                 vec.y.get_den() * vec.y.get_den()) +
                (vec.y.get_num() * vec.y.get_num() *
                 vec.x.get_den() * vec.x.get_den());

            mpz_class bit = 1;

            while (bit * bit <= sq)
            {
                bit += bit;
            }

            mpz_class rt = 0;
            while (bit > 1)
            {
                bit >>= 1;
                if ((rt + bit) * (rt + bit) <= sq)
                {
                    rt += bit;
                }
            }

            if (rt * rt != sq)
            {
                //printf("Root failed\n");
                continue;
            }

            mpq_class qrt =
                mpq_class(rt) / (vec.x.get_den() * vec.y.get_den());

            mpq_class nx = vec.x / qrt;
            mpq_class ny = vec.y / qrt;

            //printf("Root succeeded\n");

            for (uint32_t vertexIdx = 0; vertexIdx < 2; vertexIdx++)
            {
                Trf trf =
                    vertexIdx == 0 ?
                    Trf(nx, -ny,
                        ny, nx,
                        coord0.x, coord0.y) :
                    Trf(-nx, -ny,
                        -ny, nx,
                        coord1.x, coord1.y);
                Trf invTrf = invert(trf);
                    
                Shape sourceShapeBak(sourceShape);
                destUseCounts[regionId]++;

                bool addOk = addRegion(regionId, 0, trf, invTrf, vertexIdx);

                if (addOk)
                {
                    //printf("addRegion succeeded\n");
                    recurse();
                }
                else
                {
                    //printf("addRegion failed\n");
                }

                destUseCounts[regionId]--;
                sourceShape = std::move(sourceShapeBak);
            }
        }
    }
}

bool Solver::addRegion(uint32_t destRegionId,
                       uint32_t sourceRegionId,
                       const Trf& trf,
                       const Trf& invTrf,
                       uint32_t flipped)
{
    /*
    printf("    trf = (%s, %s, %s, %s) (%s, %s)\n",
           trf.a.get_str().c_str(),
           trf.b.get_str().c_str(),
           trf.c.get_str().c_str(),
           trf.d.get_str().c_str(),
           trf.x.get_str().c_str(),
           trf.y.get_str().c_str());
    printf("    invTrf = (%s, %s, %s, %s) (%s, %s)\n",
           invTrf.a.get_str().c_str(),
           invTrf.b.get_str().c_str(),
           invTrf.c.get_str().c_str(),
           invTrf.d.get_str().c_str(),
           invTrf.x.get_str().c_str(),
           invTrf.y.get_str().c_str());
    */

    auto& destRegion = destShape.m_regions[destRegionId];
    vector<Coord> destCoords(destRegion.m_edges.size());
    for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)destRegion.m_edges.size(); edgeIdx++)
    {
        auto& re = destRegion.m_edges[edgeIdx];
        uint32_t edgeId = re.m_edgeId;
        uint32_t nextVertexIdx = 1 - re.m_regionIdx;
        auto& edge = destShape.m_edges[edgeId];
        uint32_t nextVertexId = edge.m_vertices[nextVertexIdx].m_vertexId;
        auto& nextVertex = destShape.m_vertices[nextVertexId];
        destCoords[edgeIdx] = invTrf * nextVertex.m_coord;

        /*
        printf("    Vertex %" PRIu32 ": %s,%s -> %s,%s\n",
               edgeIdx,
               nextVertex.m_coord.x.get_str().c_str(),
               nextVertex.m_coord.y.get_str().c_str(),
               destCoords[edgeIdx].x.get_str().c_str(),
               destCoords[edgeIdx].y.get_str().c_str());
        */
    }

    auto& sourceRegion = sourceShape.m_regions[sourceRegionId];
    if (sourceRegion.m_type != ShapeRegionType::Interior ||
        sourceRegion.m_used != 0 ||
        sourceRegion.m_flipped != 0)
    {
        printf("Bad source region\n");
        exit(1);
    }

    // Add new vertices and split source edges
    // (only edges between an exterior region and an unused region)
    vector<uint32_t> destVertexIds;
    for (auto& coord : destCoords)
    {
        bool found = false;
        for (uint32_t vertexId = 0; vertexId < (uint32_t)sourceShape.m_vertices.size(); vertexId++)
        {
            auto& vertex = sourceShape.m_vertices[vertexId];
            if (vertex.m_coord == coord)
            {
                destVertexIds.push_back(vertexId);
                found = true;
                break;
            }
        }
        if (found)
        {
            continue;
        }

        uint32_t newVertexId = (uint32_t)sourceShape.m_vertices.size();
        destVertexIds.push_back(newVertexId);
        sourceShape.m_vertices.emplace_back();
        auto& newVertex = sourceShape.m_vertices.back();
        newVertex.m_coord = coord;

        /*
        printf("new vertex %" PRIu32 " is (%s,%s)\n",
               newVertexId,
               coord.x.get_str().c_str(),
               coord.y.get_str().c_str());
        */

        for (uint32_t edgeId = 0; edgeId < (uint32_t)sourceShape.m_edges.size(); edgeId++)
        {
            auto& edge = sourceShape.m_edges[edgeId];

            auto& prevEv = edge.m_vertices[0];
            uint32_t prevVertexId = prevEv.m_vertexId;
            auto& prevVertex = sourceShape.m_vertices[prevVertexId];
            auto& sourcePrevCoord = prevVertex.m_coord;

            auto& nextEv = edge.m_vertices[1];
            uint32_t nextVertexId = nextEv.m_vertexId;
            auto& nextVertex = sourceShape.m_vertices[nextVertexId];
            auto& sourceNextCoord = nextVertex.m_coord;

            Coord sourceVec = sourceNextCoord - sourcePrevCoord;
            Coord destVec = coord - sourcePrevCoord;
            if (cross(sourceVec, destVec) != 0)
            {
                continue;
            }
            mpq_class param =
                sourceVec.x == 0 ?
                destVec.y / sourceVec.y :
                destVec.x / sourceVec.x;
            if (param <= 0 || param >= 1)
            {
                continue;
            }

            auto& region0 = sourceShape.m_regions[edge.m_regions[0].m_regionId];
            auto& region1 = sourceShape.m_regions[edge.m_regions[1].m_regionId];
            if ((region0.m_type == ShapeRegionType::Exterior &&
                 region1.m_type == ShapeRegionType::Interior &&
                 region1.m_used == 0) ||
                (region1.m_type == ShapeRegionType::Exterior &&
                 region0.m_type == ShapeRegionType::Interior &&
                 region0.m_used == 0))
            {
                // ok
            }
            else
            {
                return false;
            }

            // Split edge

            uint32_t newEdgeId = (uint32_t)sourceShape.m_edges.size();
            //sourceShape.m_edges.emplace_back();
            //auto& newEdge = sourceShape.m_edges.back();
            ShapeEdge newEdge;

            newVertex.m_edges.resize(2);
            newVertex.m_edges[0].m_edgeId = newEdgeId;
            newVertex.m_edges[0].m_vertexIdx = 0;
            newVertex.m_edges[1].m_edgeId = edgeId;
            newVertex.m_edges[1].m_vertexIdx = 1;

            newEdge.m_vertices[0].m_vertexId = newVertexId;
            newEdge.m_vertices[0].m_edgeIdx = 0;
            newEdge.m_vertices[1].m_vertexId = nextVertexId;
            newEdge.m_vertices[1].m_edgeIdx = nextEv.m_edgeIdx;

            nextVertex.m_edges[nextEv.m_edgeIdx].m_edgeId = newEdgeId;
            nextVertex.m_edges[nextEv.m_edgeIdx].m_vertexIdx = 1;

            //edge.m_vertices[0].m_vertexId = same as before;
            //edge.m_vertices[0].m_edgeIdx = same as before;
            edge.m_vertices[1].m_vertexId = newVertexId;
            edge.m_vertices[1].m_edgeIdx = 1;

            for (uint32_t regionIdx = 0; regionIdx < 2; regionIdx++)
            {
                auto& newEr = newEdge.m_regions[regionIdx];
                auto& er = edge.m_regions[regionIdx];
                uint32_t regionId = er.m_regionId;
                newEr.m_regionId = regionId;
                uint32_t newEdgeIdx = er.m_edgeIdx;
                if (regionIdx == 0)
                {
                    newEdgeIdx++;
                }
                newEr.m_edgeIdx = newEdgeIdx;
                auto& region = sourceShape.m_regions[regionId];
                region.m_edges.emplace(region.m_edges.begin() + newEdgeIdx);
                auto& newRe = region.m_edges[newEdgeIdx];
                newRe.m_edgeId = newEdgeId;
                newRe.m_regionIdx = regionIdx;
                for (uint32_t curEdgeIdx = newEdgeIdx + 1; curEdgeIdx < (uint32_t)region.m_edges.size(); curEdgeIdx++)
                {
                    auto& curRe = region.m_edges[curEdgeIdx];
                    uint32_t curEdgeId = curRe.m_edgeId;
                    uint32_t curRegionIdx = curRe.m_regionIdx;
                    auto& curEdge = sourceShape.m_edges[curEdgeId];
                    curEdge.m_regions[curRegionIdx].m_edgeIdx++;
                }
            }

            sourceShape.m_edges.push_back(std::move(newEdge));

            break;
        }
    }

#if 1
    for (uint32_t vertexId = 0; vertexId < (uint32_t)sourceShape.m_vertices.size(); vertexId++)
    {
        auto& vertex = sourceShape.m_vertices[vertexId];
        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
        {
            auto& ve = vertex.m_edges[edgeIdx];
            uint32_t edgeId = ve.m_edgeId;
            uint32_t vertexIdx = ve.m_vertexIdx;
            auto& edge = sourceShape.m_edges[edgeId];
            auto& ev = edge.m_vertices[vertexIdx];
            if (ev.m_vertexId != vertexId ||
                ev.m_edgeIdx != edgeIdx)
            {
                printf("Corrupt sourceShape after split edges\n");
                exit(1);
            }
        }
    }
#endif

    // Map from dest edgeIdx to source edgeIdx, or -1 if not on an edge
    vector<uint32_t> destToSourceEdgeMap(destRegion.m_edges.size());

    const Coord* pDestPrevCoord = &destCoords.back();
    //for (auto& destNextCoord : destCoords)
    for (uint32_t destEdgeIdx = 0; destEdgeIdx < (uint32_t)destRegion.m_edges.size(); destEdgeIdx++)
    {
        auto& destNextCoord = destCoords[destEdgeIdx];
        auto& destPrevCoord = *pDestPrevCoord;
        pDestPrevCoord = &destNextCoord;
        Coord destDir = destNextCoord - destPrevCoord;

        destToSourceEdgeMap[destEdgeIdx] = (uint32_t)-1;

        vector<int32_t> destPrevSides(sourceRegion.m_edges.size());
        vector<int32_t> destNextSides(sourceRegion.m_edges.size());
        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)sourceRegion.m_edges.size(); edgeIdx++)
        {
            auto& re = sourceRegion.m_edges[edgeIdx];
            uint32_t edgeId = re.m_edgeId;
            uint32_t regionIdx = re.m_regionIdx;
            auto& edge = sourceShape.m_edges[edgeId];
            //auto& otherRegion = sourceShape.m_regions[edge.m_regions[1 - regionIdx].m_regionId];
            auto& sourcePrevCoord = sourceShape.m_vertices[edge.m_vertices[regionIdx].m_vertexId].m_coord;
            auto& sourceNextCoord = sourceShape.m_vertices[edge.m_vertices[1 - regionIdx].m_vertexId].m_coord;
            Coord sourceDir = sourceNextCoord - sourcePrevCoord;
            int32_t destPrevSide = sgn(cross(sourceDir, destPrevCoord - sourcePrevCoord));
            int32_t destNextSide = sgn(cross(sourceDir, destNextCoord - sourcePrevCoord));
            destPrevSides[edgeIdx] = destPrevSide;
            destNextSides[edgeIdx] = destNextSide;
        }

        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)sourceRegion.m_edges.size(); edgeIdx++)
        {
            auto& re = sourceRegion.m_edges[edgeIdx];
            uint32_t edgeId = re.m_edgeId;
            uint32_t regionIdx = re.m_regionIdx;
            auto& edge = sourceShape.m_edges[edgeId];
            //auto& otherRegion = sourceShape.m_regions[edge.m_regions[1 - regionIdx].m_regionId];
            auto& sourcePrevCoord = sourceShape.m_vertices[edge.m_vertices[regionIdx].m_vertexId].m_coord;
            auto& sourceNextCoord = sourceShape.m_vertices[edge.m_vertices[1 - regionIdx].m_vertexId].m_coord;
            Coord sourceDir = sourceNextCoord - sourcePrevCoord;
            mpq_class destParam(0);
            mpq_class sourceParam(0);
            bool intersecting = intersectLines(destPrevCoord,
                                               destDir,
                                               sourcePrevCoord,
                                               sourceDir,
                                               &destParam,
                                               &sourceParam);
            if (intersecting)
            {
                if (destParam < 0 || destParam > 1 ||
                    sourceParam < 0 || sourceParam > 1)
                {
                    // ok
                }
                else if ((destParam > 0 && destParam < 1) ||
                         (sourceParam > 0 && sourceParam < 1))
                {
                    //printf("Intersect in middle\n");
                    return false;
                }
                else if (sourceParam == 1)
                {
                    if (destParam != 0 && destParam != 1)
                    {
                        printf("Unexpected destParam\n");
                        exit(1);
                    }

                    auto& sides =
                        destParam == 0 ? destNextSides : destPrevSides;
                    if (sides[edgeIdx] < 0 &&
                        sides[(edgeIdx + 1) % sides.size()] < 0)
                    {
                        //printf("Intersect at point on wrong side\n");
                        return false;
                    }
                }
                else if (sourceParam == 0)
                {
                    // Check with the other edge
                }
                else
                {
                    printf("Unexpected sourceParam\n");
                    exit(1);
                }
            }
            else
            {
                if (cross(sourceDir, destPrevCoord - sourcePrevCoord) == 0)
                {
                    mpq_class prevParam =
                        sourceDir.x == 0 ?
                        (destPrevCoord.y - sourcePrevCoord.y) / sourceDir.y :
                        (destPrevCoord.x - sourcePrevCoord.x) / sourceDir.x;
                    mpq_class nextParam =
                        sourceDir.x == 0 ?
                        (destNextCoord.y - sourcePrevCoord.y) / sourceDir.y :
                        (destNextCoord.x - sourcePrevCoord.x) / sourceDir.x;
                    if ((prevParam <= 0 && nextParam <= 0) ||
                        (prevParam >= 1 && nextParam >= 1))
                    {
                        // ok
                    }
                    else if ((flipped == 0 && prevParam == 0 && nextParam == 1) ||
                             (flipped != 0 && prevParam == 1 && nextParam == 0))
                    {
                        // ok
                        destToSourceEdgeMap[destEdgeIdx] = edgeIdx;
                    }
                    else
                    {
                        /*
                        printf("Bad collinear configuration\n");
                        printf("prevParam = %s\n",
                               prevParam.get_str().c_str());
                        printf("nextParam = %s\n",
                               nextParam.get_str().c_str());
                        */
                        return false;
                    }
                }
            }

            // ...
        }
    }

    // Map from source edgeIdx to dest edgeIdx, or -1 if not on an edge
    vector<uint32_t> sourceToDestEdgeMap(sourceRegion.m_edges.size());
    for (uint32_t sourceEdgeIdx = 0; sourceEdgeIdx < (uint32_t)sourceRegion.m_edges.size(); sourceEdgeIdx++)
    {
        sourceToDestEdgeMap[sourceEdgeIdx] = (uint32_t)-1;
    }
    for (uint32_t destEdgeIdx = 0; destEdgeIdx < (uint32_t)destRegion.m_edges.size(); destEdgeIdx++)
    {
        uint32_t sourceEdgeIdx = destToSourceEdgeMap[destEdgeIdx];
        if (sourceEdgeIdx != (uint32_t)-1)
        {
            sourceToDestEdgeMap[sourceEdgeIdx] = destEdgeIdx;
        }
    }

    // Check that matching edges represent a fold or flat paper
    for (uint32_t destEdgeIdx = 0; destEdgeIdx < (uint32_t)destRegion.m_edges.size(); destEdgeIdx++)
    {
        uint32_t sourceEdgeIdx = destToSourceEdgeMap[destEdgeIdx];
        if (sourceEdgeIdx == (uint32_t)-1)
        {
            continue;
        }

        auto& sourceRe = sourceRegion.m_edges[sourceEdgeIdx];
        uint32_t sourceEdgeId = sourceRe.m_edgeId;
        uint32_t sourceRegionIdx = sourceRe.m_regionIdx;
        uint32_t otherRegionIdx = 1 - sourceRegionIdx;
        auto& sourceEdge = sourceShape.m_edges[sourceEdgeId];
        auto& otherEr = sourceEdge.m_regions[otherRegionIdx];
        uint32_t otherRegionId = otherEr.m_regionId;
        auto& otherRegion = sourceShape.m_regions[otherRegionId];
        if (otherRegion.m_type != ShapeRegionType::Interior ||
            otherRegion.m_used == 0)
        {
            continue;
        }

        if (flipped == otherRegion.m_flipped)
        {
            // Unmirrored case

            if (trf != otherRegion.m_trf)
            {
                //printf("Rejecting unmirrored case\n");
                return false;
            }
        }
        else
        {
            // Mirrored case

            Coord coord1 = sourceShape.m_vertices[sourceEdge.m_vertices[0].m_vertexId].m_coord;
            Coord coord2 = sourceShape.m_vertices[sourceEdge.m_vertices[1].m_vertexId].m_coord;
            Coord dir = coord2 - coord1;
            Trf mirrorTrf = getMirrorTrf(coord1, dir);
            if (trf != (otherRegion.m_trf * mirrorTrf))
            {
                //printf("Rejecting mirrored case\n");
                return false;
            }
        }
    }

    ShapeRegion usedRegion;
    usedRegion.m_type = ShapeRegionType::Interior;
    usedRegion.m_flipped = flipped;
    usedRegion.m_used = 1;
    usedRegion.m_partnerRegionId = destRegionId;
    usedRegion.m_trf = trf;
    usedRegion.m_invTrf = invTrf;
    usedRegion.m_edges.resize(destRegion.m_edges.size());
    //uint32_t usedRegionId = sourceRegionId;
    //uint32_t usedRegionId = (uint32_t)sourceShape.m_regions.size();
    uint32_t usedRegionId = (uint32_t)-2;

    // Link edges
    for (uint32_t destEdgeIdx = 0; destEdgeIdx < (uint32_t)destRegion.m_edges.size(); destEdgeIdx++)
    {
        auto& usedRe = usedRegion.m_edges[destEdgeIdx];
        uint32_t sourceEdgeIdx = destToSourceEdgeMap[destEdgeIdx];
        if (sourceEdgeIdx != (uint32_t)-1)
        {
            auto& sourceRe = sourceRegion.m_edges[sourceEdgeIdx];
            usedRe.m_edgeId = sourceRe.m_edgeId;
            usedRe.m_regionIdx = sourceRe.m_regionIdx;
            auto& sourceEdge = sourceShape.m_edges[sourceRe.m_edgeId];
            auto& sourceEr = sourceEdge.m_regions[sourceRe.m_regionIdx];
            if (sourceEr.m_regionId != sourceRegionId)
            {
                printf("sourceEr.m_regionId != sourceRegionId\n");
                exit(1);
            }
            sourceEr.m_regionId = usedRegionId;
            sourceEr.m_edgeIdx = destEdgeIdx;
        }
        else
        {
            uint32_t newEdgeId = (uint32_t)sourceShape.m_edges.size();
            ShapeEdge newEdge;
            uint32_t regionIdx = 0; // arbitrary
            usedRe.m_edgeId = newEdgeId;
            usedRe.m_regionIdx = regionIdx;
            newEdge.m_regions[regionIdx].m_regionId = usedRegionId;
            newEdge.m_regions[regionIdx].m_edgeIdx = destEdgeIdx;
            sourceShape.m_edges.push_back(std::move(newEdge));
        }
    }

    // Link vertices
    for (uint32_t prevEdgeIdx = 0; prevEdgeIdx < (uint32_t)usedRegion.m_edges.size(); prevEdgeIdx++)
    {
        uint32_t nextEdgeIdx = (prevEdgeIdx + 1) % (uint32_t)usedRegion.m_edges.size();
        /*
        printf("Linking prevIdgeIdx %" PRIu32 " to nextEdgeIdx %" PRIu32 "\n",
               prevEdgeIdx,
               nextEdgeIdx);
        */
        bool prevIsNew = (destToSourceEdgeMap[prevEdgeIdx] == (uint32_t)-1);
        bool nextIsNew = (destToSourceEdgeMap[nextEdgeIdx] == (uint32_t)-1);

#if 1
        if (!prevIsNew && !nextIsNew)
        {
            //printf("  !prevIsNew && !nextIsNew\n");
            // Nothing to do -- use existing vertex
            continue;
        }
#endif

        auto& prevRe = usedRegion.m_edges[prevEdgeIdx];
        auto& nextRe = usedRegion.m_edges[nextEdgeIdx];
        uint32_t prevEdgeId = prevRe.m_edgeId;
        uint32_t nextEdgeId = nextRe.m_edgeId;
        auto& prevEdge = sourceShape.m_edges[prevEdgeId];
        auto& nextEdge = sourceShape.m_edges[nextEdgeId];
#if 0
        printf("(%s,%s) (%s,%s) to (%s,%s) (%s,%s)\n",
               prevEdge.m_vertices[0].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[prevEdge.m_vertices[0].m_vertexId].m_coord.x.get_str().c_str(),
               prevEdge.m_vertices[0].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[prevEdge.m_vertices[0].m_vertexId].m_coord.y.get_str().c_str(),
               prevEdge.m_vertices[1].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[prevEdge.m_vertices[1].m_vertexId].m_coord.x.get_str().c_str(),
               prevEdge.m_vertices[1].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[prevEdge.m_vertices[1].m_vertexId].m_coord.y.get_str().c_str(),
               nextEdge.m_vertices[0].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[nextEdge.m_vertices[0].m_vertexId].m_coord.x.get_str().c_str(),
               nextEdge.m_vertices[0].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[nextEdge.m_vertices[0].m_vertexId].m_coord.y.get_str().c_str(),
               nextEdge.m_vertices[1].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[nextEdge.m_vertices[1].m_vertexId].m_coord.x.get_str().c_str(),
               nextEdge.m_vertices[1].m_vertexId == (uint32_t)-1 ? "_" : sourceShape.m_vertices[nextEdge.m_vertices[1].m_vertexId].m_coord.y.get_str().c_str());
        if (!prevIsNew && !nextIsNew)
        {
            //printf("  !prevIsNew && !nextIsNew\n");
            // Nothing to do -- use existing vertex
            continue;
        }
#endif
        uint32_t prevRegionIdx = prevRe.m_regionIdx;
        uint32_t nextRegionIdx = nextRe.m_regionIdx;
        uint32_t prevVertexIdx = 1 - prevRegionIdx;
        uint32_t nextVertexIdx = nextRegionIdx;
        if (flipped)
        {
            prevVertexIdx = 1 - prevVertexIdx;
            nextVertexIdx = 1 - nextVertexIdx;
        }
        auto& prevEv = prevEdge.m_vertices[prevVertexIdx];
        auto& nextEv = nextEdge.m_vertices[nextVertexIdx];

        if (prevIsNew && nextIsNew)
        {
            //printf("  prevIsNew && nextIsNew\n");
            uint32_t vertexId = destVertexIds[prevEdgeIdx];
            if (vertexId == (uint32_t)-1)
            {
                printf("vertexId is -1\n");
                exit(1);
            }
            auto& vertex = sourceShape.m_vertices[vertexId];
            /*
            printf("    (%s,%s)\n",
                   vertex.m_coord.x.get_str().c_str(),
                   vertex.m_coord.y.get_str().c_str());
            */
            if (vertex.m_edges.empty())
            {
                //printf ("    New vertex not yet used\n");
                // New vertex not yet used

                vertex.m_edges.resize(2);
                vertex.m_edges[0].m_edgeId = prevEdgeId;
                vertex.m_edges[0].m_vertexIdx = prevVertexIdx;
                vertex.m_edges[1].m_edgeId = nextEdgeId;
                vertex.m_edges[1].m_vertexIdx = nextVertexIdx;

                if (prevEv.m_vertexId != (uint32_t)-1)
                {
                    printf("prevIsNew && nextIsNew\n");
                    printf("prevEv.m_vertexId != -1\n");
                    exit(1);
                }
                if (nextEv.m_vertexId != (uint32_t)-1)
                {
                    printf("prevIsNew && nextIsNew\n");
                    printf("nextEv.m_vertexId != -1\n");
                    exit(1);
                }
                prevEv.m_vertexId = vertexId;
                prevEv.m_edgeIdx = 0;
                nextEv.m_vertexId = vertexId;
                nextEv.m_edgeIdx = 1;
            }
            else
            {
                // Vertex on source region boundary (I hope!)
                //printf("    Vertex on source region boundary (I hope!)\n");

                bool found = false;
                uint32_t foundEdgeIdx = (uint32_t)-1;

                for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
                {
                    auto& ve = vertex.m_edges[edgeIdx];
                    uint32_t edgeId = ve.m_edgeId;
                    uint32_t vertexIdx = ve.m_vertexIdx;
                    auto& edge = sourceShape.m_edges[edgeId];
                    uint32_t regionIdx = 1 - vertexIdx;
                    if (edge.m_regions[regionIdx].m_regionId == sourceRegionId)
                    {
                        if (!found)
                        {
                            found = true;
                            foundEdgeIdx = edgeIdx;
                        }
                        else
                        {
                            printf("Edge insertion index found twice\n");
                            exit(1);
                        }
                    }
                }

                if (!found)
                {
                    /*
                    printf("Edge insertion index not found\n");
                    printf("vertex.m_edges.size() = %" PRIu32 "\n",
                           (uint32_t)vertex.m_edges.size());
                    printf("Region count = %" PRIu32 "\n",
                           (uint32_t)sourceShape.m_regions.size());
                    */
                    //exit(1);
                    return false;
                }

                vertex.m_edges.emplace(vertex.m_edges.begin() + foundEdgeIdx);
                vertex.m_edges.emplace(vertex.m_edges.begin() + foundEdgeIdx);
                if (flipped == 0)
                {
                    vertex.m_edges[foundEdgeIdx].m_edgeId = nextEdgeId;
                    vertex.m_edges[foundEdgeIdx].m_vertexIdx = nextVertexIdx;
                    vertex.m_edges[foundEdgeIdx + 1].m_edgeId = prevEdgeId;
                    vertex.m_edges[foundEdgeIdx + 1].m_vertexIdx = prevVertexIdx;

                    if (prevEv.m_vertexId != (uint32_t)-1)
                    {
                        printf("prevEv.m_vertexId != -1\n");
                        exit(1);
                    }
                    if (nextEv.m_vertexId != (uint32_t)-1)
                    {
                        printf("nextEv.m_vertexId != -1\n");
                        exit(1);
                    }
                    nextEv.m_vertexId = vertexId;
                    nextEv.m_edgeIdx = foundEdgeIdx;
                    prevEv.m_vertexId = vertexId;
                    prevEv.m_edgeIdx = foundEdgeIdx + 1;
                }
                else
                {
                    vertex.m_edges[foundEdgeIdx].m_edgeId = prevEdgeId;
                    vertex.m_edges[foundEdgeIdx].m_vertexIdx = prevVertexIdx;
                    vertex.m_edges[foundEdgeIdx + 1].m_edgeId = nextEdgeId;
                    vertex.m_edges[foundEdgeIdx + 1].m_vertexIdx = nextVertexIdx;

                    if (prevEv.m_vertexId != (uint32_t)-1)
                    {
                        printf("prevEv.m_vertexId != -1\n");
                        exit(1);
                    }
                    if (nextEv.m_vertexId != (uint32_t)-1)
                    {
                        printf("nextEv.m_vertexId != -1\n");
                        exit(1);
                    }
                    prevEv.m_vertexId = vertexId;
                    prevEv.m_edgeIdx = foundEdgeIdx;
                    nextEv.m_vertexId = vertexId;
                    nextEv.m_edgeIdx = foundEdgeIdx + 1;
                }

                for (uint32_t edgeIdx = foundEdgeIdx + 2; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
                {
                    auto& ve = vertex.m_edges[edgeIdx];
                    uint32_t edgeId = ve.m_edgeId;
                    uint32_t vertexIdx = ve.m_vertexIdx;
                    auto& edge = sourceShape.m_edges[edgeId];
                    auto& ev = edge.m_vertices[vertexIdx];
                    ev.m_edgeIdx += 2;
                }
            }

            continue;
        }

        if (!prevIsNew && nextIsNew)
        {
            //printf("  !prevIsNew && nextIsNew\n");
            uint32_t vertexId = prevEv.m_vertexId;
            if (vertexId == (uint32_t)-1)
            {
                printf("vertexId is -1\n");
                exit(1);
            }
            auto& vertex = sourceShape.m_vertices[vertexId];
            /*
            printf("    (%s,%s)\n",
                   vertex.m_coord.x.get_str().c_str(),
                   vertex.m_coord.y.get_str().c_str());
            */
            uint32_t newEdgeIdx =
                flipped ?
                prevEv.m_edgeIdx + 1 :
                prevEv.m_edgeIdx;

            vertex.m_edges.emplace(vertex.m_edges.begin() + newEdgeIdx);
            vertex.m_edges[newEdgeIdx].m_edgeId = nextEdgeId;
            vertex.m_edges[newEdgeIdx].m_vertexIdx = nextVertexIdx;

            if (nextEv.m_vertexId != (uint32_t)-1)
            {
                printf("!prevIsNew && nextIsNew\n");
                printf("nextEv.m_vertexId != -1\n");
                exit(1);
            }
            nextEv.m_vertexId = vertexId;
            nextEv.m_edgeIdx = newEdgeIdx;

            for (uint32_t edgeIdx = newEdgeIdx + 1; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
            {
                auto& ve = vertex.m_edges[edgeIdx];
                uint32_t edgeId = ve.m_edgeId;
                uint32_t vertexIdx = ve.m_vertexIdx;
                auto& edge = sourceShape.m_edges[edgeId];
                auto& ev = edge.m_vertices[vertexIdx];
                ev.m_edgeIdx++;
            }

            continue;
        }

        if (prevIsNew && !nextIsNew)
        {
            //printf("  prevIsNew && !nextIsNew\n");
            uint32_t vertexId = nextEv.m_vertexId;
            if (vertexId == (uint32_t)-1)
            {
                printf("vertexId is -1\n");
                exit(1);
            }
            auto& vertex = sourceShape.m_vertices[vertexId];
            /*
            printf("    (%s,%s)\n",
                   vertex.m_coord.x.get_str().c_str(),
                   vertex.m_coord.y.get_str().c_str());
            */
            uint32_t newEdgeIdx =
                flipped ?
                nextEv.m_edgeIdx:
                nextEv.m_edgeIdx + 1;

            vertex.m_edges.emplace(vertex.m_edges.begin() + newEdgeIdx);
            vertex.m_edges[newEdgeIdx].m_edgeId = prevEdgeId;
            vertex.m_edges[newEdgeIdx].m_vertexIdx = prevVertexIdx;

            if (prevEv.m_vertexId != (uint32_t)-1)
            {
                printf("prevIsNew && !nextIsNew\n");
                printf("prevEv.m_vertexId != -1\n");
                exit(1);
            }
            prevEv.m_vertexId = vertexId;
            prevEv.m_edgeIdx = newEdgeIdx;

            for (uint32_t edgeIdx = newEdgeIdx + 1; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
            {
                auto& ve = vertex.m_edges[edgeIdx];
                uint32_t edgeId = ve.m_edgeId;
                uint32_t vertexIdx = ve.m_vertexIdx;
                auto& edge = sourceShape.m_edges[edgeId];
                auto& ev = edge.m_vertices[vertexIdx];
                ev.m_edgeIdx++;
            }

            continue;
        }
    }

#if 1
    for (uint32_t vertexId = 0; vertexId < (uint32_t)sourceShape.m_vertices.size(); vertexId++)
    {
        auto& vertex = sourceShape.m_vertices[vertexId];
        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)vertex.m_edges.size(); edgeIdx++)
        {
            auto& ve = vertex.m_edges[edgeIdx];
            uint32_t edgeId = ve.m_edgeId;
            uint32_t vertexIdx = ve.m_vertexIdx;
            auto& edge = sourceShape.m_edges[edgeId];
            auto& ev = edge.m_vertices[vertexIdx];
            if (ev.m_vertexId != vertexId ||
                ev.m_edgeIdx != edgeIdx)
            {
                printf("Corrupt sourceShape after link vertices\n");
                exit(1);
            }
        }
    }
#endif

    vector<ShapeRegion> unusedRegions;

    for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)sourceRegion.m_edges.size(); edgeIdx++)
    {
        uint32_t edgeId = sourceRegion.m_edges[edgeIdx].m_edgeId;
        uint32_t regionIdx = sourceRegion.m_edges[edgeIdx].m_regionIdx;
        auto& edge = sourceShape.m_edges[edgeId];
        if (edge.m_regions[regionIdx].m_regionId != sourceRegionId)
        {
            /*
            printf("Continuing (regionId = %" PRId32 ", other is %" PRId32 ")\n",
                   edge.m_regions[regionIdx].m_regionId,
                   edge.m_regions[1 - regionIdx].m_regionId);
            */
            continue;
        }
        /*
        printf("Processing (regionId = %" PRId32 ", other is %" PRId32 ")\n",
               edge.m_regions[regionIdx].m_regionId,
               edge.m_regions[1 - regionIdx].m_regionId);
        */

        uint32_t newRegionId =
            (uint32_t)sourceShape.m_regions.size() +
            (uint32_t)unusedRegions.size();
        unusedRegions.emplace_back();
        auto& newRegion = unusedRegions.back();
        newRegion.m_type = ShapeRegionType::Interior;
        newRegion.m_flipped = 0;
        newRegion.m_used = 0;

        uint32_t curEdgeId = edgeId;
        uint32_t curRegionIdx = regionIdx;
        do
        {
            uint32_t newEdgeIdx = (uint32_t)newRegion.m_edges.size();
            newRegion.m_edges.emplace_back();
            auto& newRe = newRegion.m_edges.back();
            newRe.m_edgeId = curEdgeId;
            newRe.m_regionIdx = curRegionIdx;

            auto& curEdge = sourceShape.m_edges[curEdgeId];
            auto& curEr = curEdge.m_regions[curRegionIdx];
            if (curEr.m_regionId != sourceRegionId &&
                curEr.m_regionId != (uint32_t)-1)
            {
                printf("curEr.m_regionId != sourceRegionId\n");
                printf("curEr.m_regionId == %" PRId32 " (other is %" PRId32 ")\n",
                       curEr.m_regionId,
                       curEdge.m_regions[1 - curRegionIdx].m_regionId);
                printf("(%s,%s) (%s,%s)\n",
                       sourceShape.m_vertices[curEdge.m_vertices[0].m_vertexId].m_coord.x.get_str().c_str(),
                       sourceShape.m_vertices[curEdge.m_vertices[0].m_vertexId].m_coord.y.get_str().c_str(),
                       sourceShape.m_vertices[curEdge.m_vertices[1].m_vertexId].m_coord.x.get_str().c_str(),
                       sourceShape.m_vertices[curEdge.m_vertices[1].m_vertexId].m_coord.y.get_str().c_str());
                throw SolverErrorException();
                exit(1);
            }
            curEr.m_regionId = newRegionId;
            curEr.m_edgeIdx = newEdgeIdx;

            uint32_t nextVertexIdx = 1 - curRegionIdx;
            auto& nextEv = curEdge.m_vertices[nextVertexIdx];
            if (nextEv.m_vertexId >= (uint32_t)sourceShape.m_vertices.size())
            {
                printf("nextEv.m_vertexId %" PRIu32 " > size %" PRIu32 "\n",
                       nextEv.m_vertexId,
                       (uint32_t)sourceShape.m_vertices.size());
            }
            auto& nextVertex = sourceShape.m_vertices[nextEv.m_vertexId];
            auto& nextVe = nextVertex.m_edges[(nextEv.m_edgeIdx + (uint32_t)nextVertex.m_edges.size() - 1) % (uint32_t)nextVertex.m_edges.size()];

            curEdgeId = nextVe.m_edgeId;
            curRegionIdx = nextVe.m_vertexIdx;
        }
        while (curEdgeId != edgeId);
    }

    for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)usedRegion.m_edges.size(); edgeIdx++)
    {
        auto& re = usedRegion.m_edges[edgeIdx];
        uint32_t edgeId = re.m_edgeId;
        uint32_t regionIdx = re.m_regionIdx;
        auto& edge = sourceShape.m_edges[edgeId];
        auto& er = edge.m_regions[regionIdx];
        if (er.m_edgeIdx != edgeIdx)
        {
            printf("er.m_edgeIdx != edgeIdx\n");
            exit(1);
        }
        er.m_regionId = sourceRegionId;
    }

    sourceShape.m_regions[sourceRegionId] = std::move(usedRegion);

    for (auto& unusedRegion : unusedRegions)
    {
        sourceShape.m_regions.push_back(std::move(unusedRegion));
    }

    return true;
}

#define USE_CHOICES 1

#if USE_CHOICES

class Option
{
public:
    Option() :
        m_destRegionId((uint32_t)-1),
        m_flipped(0)
    {
    }

    uint32_t m_destRegionId;
    Trf m_trf;
    Trf m_invTrf;
    uint32_t m_flipped;
};

class Choice
{
public:
    Choice() :
        m_unusedRegionId((uint32_t)-1),
        m_usedRegionId((uint32_t)-1),
        m_usedEdgeIdx((uint32_t)-1)
    {
    }

    uint32_t m_unusedRegionId;
    uint32_t m_usedRegionId;
    uint32_t m_usedEdgeIdx;
    vector<Option> m_options;
};

void Solver::recurse()
{
    m_nodeCount++;
    if (m_stopTimeMS != 0 &&
        getTimeMS() > m_stopTimeMS)
    {
        throw SolverTimeoutException();
    }

    vector<uint32_t> unusedRegionIds;

    uint32_t sourceUnusedCount = 0;
    for (uint32_t sourceRegionId = 0; sourceRegionId < (uint32_t)sourceShape.m_regions.size(); sourceRegionId++)
    {
        auto& sourceRegion = sourceShape.m_regions[sourceRegionId];
        if (sourceRegion.m_type == ShapeRegionType::Interior &&
            sourceRegion.m_used == 0)
        {
            sourceUnusedCount++;
            unusedRegionIds.push_back(sourceRegionId);
        }
    }

    uint32_t destUnusedCount = 0;
    for (uint32_t destRegionId = 0; destRegionId < (uint32_t)destShape.m_regions.size(); destRegionId++)
    {
        auto& destRegion = destShape.m_regions[destRegionId];
        if (destRegion.m_type == ShapeRegionType::Interior &&
            destUseCounts[destRegionId] == 0)
        {
            destUnusedCount++;
        }
    }

    if (sourceUnusedCount == 0)
    {
        if (destUnusedCount == 0)
        {
            prepareSolution();
        }
        return;
    }

    vector<Choice> choices;

    for (uint32_t unusedRegionId : unusedRegionIds)
    {
        auto& unusedRegion = sourceShape.m_regions[unusedRegionId];

        for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)unusedRegion.m_edges.size(); edgeIdx++)
        {
            auto& re = unusedRegion.m_edges[edgeIdx];
            uint32_t edgeId = re.m_edgeId;
            uint32_t regionIdx = re.m_regionIdx;
            uint32_t otherRegionIdx = 1 - regionIdx;
            auto& edge = sourceShape.m_edges[edgeId];
            auto& otherEr = edge.m_regions[otherRegionIdx];
            uint32_t otherRegionId = otherEr.m_regionId;
            uint32_t otherEdgeIdx = otherEr.m_edgeIdx;
            auto& otherRegion = sourceShape.m_regions[otherRegionId];
            if (otherRegion.m_type == ShapeRegionType::Interior &&
                otherRegion.m_used != 0)
            {
                choices.emplace_back();
                auto& choice = choices.back();
                choice.m_unusedRegionId = unusedRegionId;
                choice.m_usedRegionId = otherRegionId;
                choice.m_usedEdgeIdx = otherEdgeIdx;
            }
        }
    }

    if (choices.empty())
    {
        printf("No choices found\n");
        exit(1);
    }

    uint32_t bestOptionCount = (uint32_t)-1;

    for (auto& choice : choices)
    {
        uint32_t unusedRegionId = choice.m_unusedRegionId;
        uint32_t usedRegionId = choice.m_usedRegionId;
        uint32_t usedEdgeIdx = choice.m_usedEdgeIdx;

        for (uint32_t mirrored = 0; mirrored < 2; mirrored++)
        {
            if (m_stopTimeMS != 0 &&
                getTimeMS() > m_stopTimeMS)
            {
                throw SolverTimeoutException();
            }

            auto& usedRegion = sourceShape.m_regions[usedRegionId];
            uint32_t destUsedRegionId = usedRegion.m_partnerRegionId;
            if (destUsedRegionId >= (uint32_t)destShape.m_regions.size())
            {
                printf("Bad partner region ID\n");
                exit(1);
            }
            auto& destUsedRegion = destShape.m_regions[destUsedRegionId];

            uint32_t destRegionId = (uint32_t)-1;
            Trf trf;
            Trf invTrf;
            uint32_t flipped = 0;

            if (mirrored == 0)
            {
                // Unmirrored

                auto& destUsedRe = destUsedRegion.m_edges[usedEdgeIdx];
                uint32_t destEdgeId = destUsedRe.m_edgeId;
                uint32_t destRegionIdx = destUsedRe.m_regionIdx;
                uint32_t destOtherRegionIdx = 1 - destRegionIdx;
                auto& destEdge = destShape.m_edges[destEdgeId];
                auto& destOtherEr = destEdge.m_regions[destOtherRegionIdx];
                uint32_t destOtherRegionId = destOtherEr.m_regionId;
                auto& destOtherRegion = destShape.m_regions[destOtherRegionId];
                if (destOtherRegion.m_type != ShapeRegionType::Interior)
                {
                    continue;
                }

                destRegionId = destOtherRegionId;
                trf = usedRegion.m_trf;
                invTrf = usedRegion.m_invTrf;
                flipped = usedRegion.m_flipped;
            }
            else
            {
                // Mirrored

                destRegionId = destUsedRegionId;
                flipped = 1 - usedRegion.m_flipped;

                auto& usedRe = usedRegion.m_edges[usedEdgeIdx];
                uint32_t usedEdgeId = usedRe.m_edgeId;
                auto& usedEdge = sourceShape.m_edges[usedEdgeId];
                Coord coord1 = sourceShape.m_vertices[usedEdge.m_vertices[0].m_vertexId].m_coord;
                Coord coord2 = sourceShape.m_vertices[usedEdge.m_vertices[1].m_vertexId].m_coord;
                Coord dir = coord2 - coord1;
                Trf mirrorTrf = getMirrorTrf(coord1, dir);
                trf = usedRegion.m_trf * mirrorTrf;
                invTrf = invert(trf);
            }
        
            Shape sourceShapeBak(sourceShape);
            destUseCounts[destRegionId]++;

            bool addOk = addRegion(destRegionId,
                                   unusedRegionId,
                                   trf,
                                   invTrf,
                                   flipped);

            if (addOk)
            {
                choice.m_options.emplace_back();
                auto& option = choice.m_options.back();
                option.m_destRegionId = destRegionId;
                option.m_trf = trf;
                option.m_invTrf = invTrf;
                option.m_flipped = flipped;
                //recurse();
            }

            destUseCounts[destRegionId]--;
            sourceShape = std::move(sourceShapeBak);
        }

        uint32_t optionCount = (uint32_t)choice.m_options.size();
        if (optionCount == 0)
        {
            //printf("bestOptionCount = 0\n");
            return;
        }

        if (bestOptionCount == (uint32_t)-1 ||
            optionCount < bestOptionCount)
        {
            bestOptionCount = optionCount;
        }
    }

    //printf("bestOptionCount = %" PRIu32 "\n", bestOptionCount);

    vector<uint32_t> choiceIds;
    for (uint32_t choiceId = 0; choiceId < (uint32_t)choices.size(); choiceId++)
    {
        if ((uint32_t)choices[choiceId].m_options.size() == bestOptionCount)
        {
            choiceIds.push_back(choiceId);
        }
    }

    uint32_t choiceIdIdx = (uint32_t)(m_random.get() % choiceIds.size());
    uint32_t choiceId = choiceIds[choiceIdIdx];
    auto& choice = choices[choiceId];

    if ((uint32_t)choice.m_options.size() != bestOptionCount)
    {
        printf("bestOptionCount mismatch\n");
        exit(1);
    }

    for (auto& option : choice.m_options)
    {
        uint32_t destRegionId = option.m_destRegionId;

        Shape sourceShapeBak(sourceShape);
        destUseCounts[destRegionId]++;

        bool addOk = addRegion(destRegionId,
                               choice.m_unusedRegionId,
                               option.m_trf,
                               option.m_invTrf,
                               option.m_flipped);

        if (addOk)
        {
            recurse();
        }
        else
        {
            printf("addOk expected\n");
            exit(1);
        }

        destUseCounts[destRegionId]--;
        sourceShape = std::move(sourceShapeBak);
    }
}

#else

void Solver::recurse()
{
    m_nodeCount++;
    if (m_stopTimeMS != 0 &&
        getTimeMS() > m_stopTimeMS)
    {
        throw SolverTimeoutException();
    }

    uint32_t unusedRegionId = (uint32_t)-1;

    uint32_t sourceUnusedCount = 0;
    for (uint32_t sourceRegionId = 0; sourceRegionId < (uint32_t)sourceShape.m_regions.size(); sourceRegionId++)
    {
        auto& sourceRegion = sourceShape.m_regions[sourceRegionId];
        if (sourceRegion.m_type == ShapeRegionType::Interior &&
            sourceRegion.m_used == 0)
        {
            sourceUnusedCount++;
            if (unusedRegionId == (uint32_t)-1)
            {
                unusedRegionId = sourceRegionId;
            }
        }
    }

    uint32_t destUnusedCount = 0;
    for (uint32_t destRegionId = 0; destRegionId < (uint32_t)destShape.m_regions.size(); destRegionId++)
    {
        auto& destRegion = destShape.m_regions[destRegionId];
        if (destRegion.m_type == ShapeRegionType::Interior &&
            destUseCounts[destRegionId] == 0)
        {
            destUnusedCount++;
        }
    }

    /*
    printf("sourceUnusedCount = %" PRIu32 ", destUnusedCount = %" PRIu32 "\n",
           sourceUnusedCount,
           destUnusedCount);
    */

    if (sourceUnusedCount == 0)
    {
        if (destUnusedCount == 0)
        {
            //printf("SOLVED!\n");
            prepareSolution();
        }
        return;
    }

    auto& unusedRegion = sourceShape.m_regions[unusedRegionId];
    uint32_t usedRegionId = (uint32_t)-1;
    uint32_t usedEdgeIdx = (uint32_t)-1;


    vector<uint32_t> edgeIdxs(unusedRegion.m_edges.size());
    for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)unusedRegion.m_edges.size(); edgeIdx++)
    {
        edgeIdxs[edgeIdx] = edgeIdx;
    }
#if 1
    {
        uint32_t idx = (uint32_t)unusedRegion.m_edges.size() - 1;
        while (idx > 0)
        {
            using std::swap;
            swap(edgeIdxs[idx], edgeIdxs[m_random.get() % (idx + 1)]);
            idx--;
        }
    }
#endif

    //for (uint32_t edgeIdx = 0; edgeIdx < (uint32_t)unusedRegion.m_edges.size(); edgeIdx++)
    for (uint32_t edgeIdx : edgeIdxs)
    {
        auto& re = unusedRegion.m_edges[edgeIdx];
        uint32_t edgeId = re.m_edgeId;
        uint32_t regionIdx = re.m_regionIdx;
        uint32_t otherRegionIdx = 1 - regionIdx;
        auto& edge = sourceShape.m_edges[edgeId];
        auto& otherEr = edge.m_regions[otherRegionIdx];
        uint32_t otherRegionId = otherEr.m_regionId;
        uint32_t otherEdgeIdx = otherEr.m_edgeIdx;
        auto& otherRegion = sourceShape.m_regions[otherRegionId];
        if (otherRegion.m_type == ShapeRegionType::Interior &&
            otherRegion.m_used != 0)
        {
            usedRegionId = otherRegionId;
            usedEdgeIdx = otherEdgeIdx;
            break;
        }
    }

    if (usedRegionId == (uint32_t)-1 ||
        usedEdgeIdx == (uint32_t)-1)
    {
        printf("Used region not found\n");
        exit(1);
    }

    for (uint32_t mirrored = 0; mirrored < 2; mirrored++)
    {
        /*
        if (m_stopTimeMS != 0 &&
            getTimeMS() > m_stopTimeMS)
        {
            throw SolverTimeoutException();
        }
        */

        auto& usedRegion = sourceShape.m_regions[usedRegionId];
        uint32_t destUsedRegionId = usedRegion.m_partnerRegionId;
        if (destUsedRegionId >= (uint32_t)destShape.m_regions.size())
        {
            printf("Bad partner region ID\n");
            exit(1);
        }
        auto& destUsedRegion = destShape.m_regions[destUsedRegionId];

        uint32_t destRegionId = (uint32_t)-1;
        Trf trf;
        Trf invTrf;
        uint32_t flipped = 0;

        if (mirrored == 0)
        {
            // Unmirrored

            auto& destUsedRe = destUsedRegion.m_edges[usedEdgeIdx];
            uint32_t destEdgeId = destUsedRe.m_edgeId;
            uint32_t destRegionIdx = destUsedRe.m_regionIdx;
            uint32_t destOtherRegionIdx = 1 - destRegionIdx;
            auto& destEdge = destShape.m_edges[destEdgeId];
            auto& destOtherEr = destEdge.m_regions[destOtherRegionIdx];
            uint32_t destOtherRegionId = destOtherEr.m_regionId;
            auto& destOtherRegion = destShape.m_regions[destOtherRegionId];
            if (destOtherRegion.m_type != ShapeRegionType::Interior)
            {
                continue;
            }

            destRegionId = destOtherRegionId;
            trf = usedRegion.m_trf;
            invTrf = usedRegion.m_invTrf;
            flipped = usedRegion.m_flipped;
        }
        else
        {
            // Mirrored

            destRegionId = destUsedRegionId;
            flipped = 1 - usedRegion.m_flipped;

            auto& usedRe = usedRegion.m_edges[usedEdgeIdx];
            uint32_t usedEdgeId = usedRe.m_edgeId;
            auto& usedEdge = sourceShape.m_edges[usedEdgeId];
            Coord coord1 = sourceShape.m_vertices[usedEdge.m_vertices[0].m_vertexId].m_coord;
            Coord coord2 = sourceShape.m_vertices[usedEdge.m_vertices[1].m_vertexId].m_coord;
            Coord dir = coord2 - coord1;
            Trf mirrorTrf = getMirrorTrf(coord1, dir);
            trf = usedRegion.m_trf * mirrorTrf;
            invTrf = invert(trf);
        }

        Shape sourceShapeBak(sourceShape);
        destUseCounts[destRegionId]++;

        bool addOk = addRegion(destRegionId,
                               unusedRegionId,
                               trf,
                               invTrf,
                               flipped);

        if (addOk)
        {
            //printf("addRegion succeeded\n");
            recurse();
        }
        else
        {
            //printf("addRegion failed\n");
        }

        destUseCounts[destRegionId]--;
        sourceShape = std::move(sourceShapeBak);
    }
}

#endif

void Solver::prepareSolution()
{
    vector<uint32_t> okEdges[2];
    okEdges[0].resize(sourceShape.m_edges.size());
    okEdges[1].resize(sourceShape.m_edges.size());

    for (uint32_t dir = 0; dir < 2; dir++)
    {
        for (uint32_t edgeId = 0; edgeId < (uint32_t)sourceShape.m_edges.size(); edgeId++)
        {
            auto& edge = sourceShape.m_edges[edgeId];
            auto& insideEr = edge.m_regions[dir];
            auto& outsideEr = edge.m_regions[1 - dir];
            auto& insideRegion = sourceShape.m_regions[insideEr.m_regionId];
            auto& outsideRegion = sourceShape.m_regions[outsideEr.m_regionId];

            if (insideRegion.m_type != ShapeRegionType::Interior)
            {
                continue;
            }

            if (outsideRegion.m_type == ShapeRegionType::Interior &&
                insideRegion.m_flipped == outsideRegion.m_flipped)
            {
                continue;
            }

            okEdges[dir][edgeId] = 1;
            /*
            printf("okEdges[%" PRIu32 "][%" PRIu32 "] = 1\n",
                   dir, edgeId);
            */
        }
    }

    vector<uint32_t> okVertices(sourceShape.m_vertices.size());
    for (uint32_t vertexId = 0; vertexId < (uint32_t)sourceShape.m_vertices.size(); vertexId++)
    {
        vector<uint32_t> okEdgeIds;

        auto& vertex = sourceShape.m_vertices[vertexId];
        for (auto& ve : vertex.m_edges)
        {
            uint32_t edgeId = ve.m_edgeId;
            if (okEdges[0][edgeId] || okEdges[1][edgeId])
            {
                if (std::find(okEdgeIds.begin(), okEdgeIds.end(), edgeId) ==
                    okEdgeIds.end())
                {
                    okEdgeIds.push_back(edgeId);
                }
            }
        }

        auto& okVertex = okVertices[vertexId];

        if (okEdgeIds.size() > 2)
        {
            okVertex = 1;
            continue;
        }

        if (okEdgeIds.size() < 2)
        {
            if (okEdgeIds.size() == 1)
            {
                printf("okEdgeIds.size() == 1\n");
                exit(1);
            }
            okVertex = 0;
            continue;
        }

        auto& edge0 = sourceShape.m_edges[okEdgeIds[0]];
        auto& edge1 = sourceShape.m_edges[okEdgeIds[1]];
        auto& coord0_0 = sourceShape.m_vertices[edge0.m_vertices[0].m_vertexId].m_coord;
        auto& coord0_1 = sourceShape.m_vertices[edge0.m_vertices[1].m_vertexId].m_coord;
        auto& coord1_0 = sourceShape.m_vertices[edge1.m_vertices[0].m_vertexId].m_coord;
        auto& coord1_1 = sourceShape.m_vertices[edge1.m_vertices[1].m_vertexId].m_coord;
        Coord vec0 = coord0_1 - coord0_0;
        Coord vec1 = coord1_1 - coord1_0;

        if (cross(vec0, vec1) == 0)
        {
            okVertex = 0;
        }
        else
        {
            okVertex = 1;
        }
    }

    vector<uint32_t> processedEdges[2];
    processedEdges[0].resize(sourceShape.m_edges.size());
    processedEdges[1].resize(sourceShape.m_edges.size());

    vector<uint32_t> vertexMap(sourceShape.m_vertices.size(), (uint32_t)-1);

    Solution solution;

    for (uint32_t dir = 0; dir < 2; dir++)
    {
        for (uint32_t edgeId = 0; edgeId < (uint32_t)sourceShape.m_edges.size(); edgeId++)
        {
            if (okEdges[dir][edgeId] == 0 ||
                processedEdges[dir][edgeId] != 0)
            {
                continue;
            }

            vector<uint32_t> facet;

            uint32_t curDir = dir;
            uint32_t curEdgeId = edgeId;
            do
            {
                processedEdges[curDir][curEdgeId] = 1;
                /*
                printf("processedEdges[%" PRIu32 "][%" PRIu32 "] = 1\n",
                       curDir, curEdgeId);
                */

                auto& edge = sourceShape.m_edges[curEdgeId];
                auto& ev = edge.m_vertices[1 - curDir];
                uint32_t vertexId = ev.m_vertexId;
                uint32_t edgeIdx = ev.m_edgeIdx;
                auto& vertex = sourceShape.m_vertices[vertexId];

                if (okVertices[vertexId] != 0)
                {
                    if (vertexMap[vertexId] == (uint32_t)-1)
                    {
                        vertexMap[vertexId] = (uint32_t)solution.m_sourceCoords.size();
                        solution.m_sourceCoords.push_back(vertex.m_coord);

                        auto& er = edge.m_regions[curDir];
                        uint32_t regionId = er.m_regionId;
                        auto& region = sourceShape.m_regions[regionId];
                        solution.m_destCoords.push_back(region.m_trf * vertex.m_coord);
                    }
                    facet.push_back(vertexMap[vertexId]);
                }

                do
                {
                    edgeIdx = (edgeIdx + (uint32_t)vertex.m_edges.size() - 1) % (uint32_t)vertex.m_edges.size();
                    auto& ve = vertex.m_edges[edgeIdx];
                    curEdgeId = ve.m_edgeId;
                    curDir = ve.m_vertexIdx;
                }
                while (okEdges[curDir][curEdgeId] == 0);
            }
            while (curDir != dir || curEdgeId != edgeId);

            solution.m_facets.push_back(std::move(facet));
        }
    }

    Solution smallestSolution;
    string smallestSolutionData;
    uint32_t smallestSize = 0;

    for (uint32_t bits = 0; bits < 8; bits++)
    {
        Solution curSolution(solution);
        for (auto& coord : curSolution.m_sourceCoords)
        {
            if (bits & 1)
            {
                coord.x = 1 - coord.x;
            }
            if (bits & 2)
            {
                coord.y = 1 - coord.y;
            }
            if (bits & 4)
            {
                using std::swap;
                swap(coord.x, coord.x);
            }
        }

        uint32_t curSize = 0;
        string msg;
        string curSolutionData;
        if (!formatSolution(curSolution,
                            &curSolutionData,
                            &curSize,
                            &msg))
        {
            printf("formatSolution: %s\n", msg.c_str());
            continue;
        }

        //printf("%s", curSolutionData.c_str());

        /*
        printf("Solution %" PRIu32 " size: %" PRIu32 "\n",
               bits,
               curSize);
        */

        if (bits == 0 ||
            curSize < smallestSize)
        {
            smallestSolution = std::move(curSolution);
            smallestSolutionData = std::move(curSolutionData);
            smallestSize = curSize;
        }
    }

    printf("SOLVED!  size = %" PRIu32 "\n", smallestSize);

    if (!m_solved || smallestSize < m_solutionSize)
    {
        m_solved = true;
        m_solution = std::move(smallestSolution);
        m_solutionData = std::move(smallestSolutionData);
        m_solutionSize = smallestSize;
    }

    throw SolverDoneException();
}

bool solve(const Problem& problem,
           uint64_t seed,
           uint32_t maxDurationMS,
           Solution* pSolution,
           string* pSolutionData,
           uint32_t* pSolutionSize,
           uint64_t* pNodeCount)
{
    Solver solver(seed);
    try
    {
        solver.solve(problem, maxDurationMS);
    }
    catch (const SolverDoneException&)
    {
    }
    catch (const SolverTimeoutException&)
    {
    }
    catch (const SolverErrorException&)
    {
    }

    if (solver.m_solved)
    {
        if (pSolution)
        {
            *pSolution = std::move(solver.m_solution);
        }
        if (pSolutionData)
        {
            *pSolutionData = std::move(solver.m_solutionData);
        }
        if (pSolutionSize)
        {
            *pSolutionSize = solver.m_solutionSize;
        }
    }

    if (pNodeCount)
    {
        *pNodeCount = solver.m_nodeCount;
    }

    return solver.m_solved;
}
